#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-10-12
# File  : nvdct/lib/args.py

#
# options used
# -b --backend
# -c --config
# -d --default
# -l --layers
# -o --output-directory
# -v --version
# --api-port (deprecated ?)
# --check-config-only
# --dont-compare
# --filter-customers
# --filter-sites
# --keep-max-topologies
# --l2-case
# --l2-display-ports
# --l2-display-neighbours
# --l2-ignore-mismatch
# --l2-prefix
# --l2-remove-domain
# --l2-skip-external
# --l3-display-devices
# --l3-include-hosts
# --l3-include-loopback
# --l3-skip-cidr-0
# --l3-skip-cidr-32-128
# --l3-skip-if
# --l3-skip-ip
# --l3-skip-public
# --log-file
# --log-level
# --log-to-stdout
# --min-topology-age
# --pre-fetch
# --quiet
# --time-format
# --update-config


from argparse import (
    ArgumentParser,
    Namespace as arg_Namespace,
    RawTextHelpFormatter,
)
from pathlib import Path

from lib.constants import (
    Backends,
    CONFIG_FILE,
    CliLong,
    CliShort,
    ExitCodes,
    IncludeExclude,
    L2Case,
    L2IgnoreMismatch,
    L2RemoveDomain,
    Layers,
    LogLevels,
    MinVersions,
    NVDCT_VERSION,
    SCRIPT,
    TIME_FORMAT_ARGPARSER,
    TomlSections,
    URLs,
)


def parse_arguments() -> arg_Namespace:
    parser = ArgumentParser(
        prog='nvdct.py',
        description=(
            'This script creates the topology data file needed for the Checkmk Network Visualization,\n'  # noqa: E501
            'For more information see the announcement from schnetz in the Checkmk forum:\n'
            f'{URLs.FORUM_SCHNETZ}\n'
            '\n'
            'The required plugins to create the inventory data can be found here:\n'
            f'{URLs.TOPIC_NV}\n'
            '\n'
            f'\nVersion: {NVDCT_VERSION} | Written by: thl-cmk\n'
            f'for more information see: {URLs.NVDCT}'
        ),
        formatter_class=RawTextHelpFormatter,
        epilog='Exit codes:\n'
               f' {ExitCodes.OK} - No error\n'
               f' {ExitCodes.BAD_OPTION_LIST} - Bad options list\n'
               f' {ExitCodes.BAD_TOML_FORMAT} - Bad TOML file format\n'
               f' {ExitCodes.BACKEND_NOT_IMPLEMENTED} - Backend not implemented\n'
               f' {ExitCodes.AUTOMATION_SECRET_NOT_FOUND} - Automation secret not found\n'
               f' {ExitCodes.NO_LAYER_CONFIGURED} - No layer to work on\n'
               '\nUsage:\n'
               f'{SCRIPT} -c ~/local/bin/nvdct/conf/my_{CONFIG_FILE} \n\n'
    )
    parser.add_argument(
        CliShort.BACKEND, CliLong.BACKEND,
        choices=Backends.list(),
        # default='MULTISITE',
        help='Backend used to retrieve the topology data\n'
             f' - {Backends.LIVESTATUS} : fetches data via local Livestatus (local site only)\n'
             f' - {Backends.MULTISITE}  : like LIVESTATUS but for distributed environments (default)\n'
             f' - {Backends.RESTAPI}    : uses the Checkmk REST API.',
    )
    parser.add_argument(
        CliShort.CONFIG, CliLong.CONFIG, type=str,
        help='Set the name of the config file.\n'
             f'Default is ~/local/bin/nvdct/conf/{CONFIG_FILE}',
    )
    parser.add_argument(
        CliShort.DEFAULT, CliLong.DEFAULT, action='store_const', const=True,  # default=False,
        help='Set the created topology data as default. Will be created automatically\n'
             'if it doesnt exists.',
    )
    parser.add_argument(
        CliShort.LAYERS, CliLong.LAYERS,
        nargs='+',
        choices=Layers.list(),
        help=(
            f' - {Layers.CDP}    : needs inv_cdp_cache package at least in version {MinVersions.CDP}\n'
            f' - {Layers.LLDP}   : needs inv_lldp_cache package at least in version {MinVersions.LLDP}\n'
            f' - {Layers.L3V4}   : needs inv_ip_address package at least in version {MinVersions.SNMP_IP_ADDRESSES} for SNMP based hosts\n'
            f'            for Linux based hosts inv_lnx_ip_if in version {MinVersions.LINUX_IP_ADDRESSES}\n'
            f'            for Windows based hosts inv_win_ip_if in version {MinVersions.WINDOWS_IP_ADDRESSES}\n'
            f' - {Layers.STATIC} : creates a topology base on the "[{TomlSections.STATIC_CONNECTIONS}]" section in the TOML file\n'
        )
    )
    parser.add_argument(
        CliShort.OUTPUT_DIRECTORY, CliLong.OUTPUT_DIRECTORY, type=str,
        help='Directory name where to save the topology data.\n'
             'I.e.: my_topology. Default is the actual date/time\n'
             f'in "{CliLong.TIME_FORMAT}" format.\n'
             'NOTE: the directory is a sub directory under "~/var/check_mk/topology/data/"\n',
    )
    parser.add_argument(
        CliShort.VERSION, CliLong.VERSION, action='version',
        version=f'{Path(SCRIPT).name} version: {NVDCT_VERSION}',
        help='Print version of this script and exit',
    )
    parser.add_argument(
        CliLong.API_PORT, type=int,  # default=False,
        help='TCP Port to access the REST API. By NVDCT will try to automatically\n'
             'detect the site apache port.',
    )
    parser.add_argument(
        CliLong.CHECK_CONFIG, action='store_const', const=True,  # default=False,
        help=f'Only tries to read/parse the config file from {CONFIG_FILE} and exits.',
    )
    parser.add_argument(
        CliLong.DONT_COMPARE, action='store_const', const=True,  # default=False,
        help='Do not compare the actual topology data with the default topology\n'
             'data. By default, the actual topology is compared with the default\n'
             'topology. If the data matches, the actual topology is not saved.\n'
             'So, if you run this tool in a cron job, a new topology will be\n'
             f'created only if there was a change, unless you use "{CliLong.DONT_COMPARE}".'
    )
    parser.add_argument(
        CliLong.FILTER_CUSTOMERS,
        choices=IncludeExclude.list(),
        # default='INCLUDE',
        help=f'{IncludeExclude.INCLUDE}/{IncludeExclude.EXCLUDE} customer list "[{TomlSections.FILTER_BY_CUSTOMER}]" from TOML file.'
             f'NOTE: {Backends.MULTISITE} backend only.',
    )
    parser.add_argument(
        CliLong.FILTER_SITES,
        choices=IncludeExclude.list(),
        # default='INCLUDE',
        help=f'{IncludeExclude.INCLUDE}/{IncludeExclude.EXCLUDE} site list "[{TomlSections.FILTER_BY_SITE}]" from TOML file.'
    )
    parser.add_argument(
        CliLong.KEEP_MAX_TOPOLOGIES, type=int,
        help='Number of topologies to keep. The oldest topologies above keep\n'
             'will be deleted.\n'
             'NOTE: The default/protected topologies will be kept always.'
    )
    parser.add_argument(
        CliLong.L2_CASE,
        choices=L2Case.list(),
        help='Change L2 neighbour name case before matching to Checkmk host\n'
             f'- {L2Case.OFF}         : Do not change the case of the neighbour name\n'
             f'- {L2Case.INSENSITIVE} : search for a matching host by ignoring the case of neighbour name and host name\n'
             f'- {L2Case.LOWER}       : change to all lower case\n'
             f'- {L2Case.UPPER}       : change to all upper case, without the domain part of the neighbour name\n'
             f'- {L2Case.AUTO}        : try all the above variants\n'
             f'Default is let the case of the neighbour name untouched ("OFF").\n'
             f'Takes place after "{CliLong.L2_REMOVE_DOMAIN}" and before "{CliLong.L2_PREFIX}"',
    )
    parser.add_argument(
        CliLong.L2_DISPLAY_PORTS, action='store_const', const=True,  # default=False,
        help='Use L2 port names as display name for interfaces in L2 topologies',
    )
    parser.add_argument(
        CliLong.L2_DISPLAY_NEIGHBOURS, action='store_const', const=True,  # default=False,
        help='Use L2 neighbour name as display name in L2 topologies',
    )
    parser.add_argument(
        CliLong.L2_IGNORE_MISMATCH,
        nargs='+',
        choices=L2IgnoreMismatch.list(),
        help=(
            f' - {L2IgnoreMismatch.DUPLEX} : Ignore duplex mismatch in layer 2 topologies\n'
            f' - {L2IgnoreMismatch.SPEED}  : Ignore speed mismatch in layer 2 topologies\n'
            f' - {L2IgnoreMismatch.VLAN}   : Ignore native vlan mismatch in layer 2 topologies\n'
        )
    )
    parser.add_argument(
        CliLong.L2_PREFIX, type=str,
        help=f'Prepends each L2 neighbour name with the prefix before matching to a Checkmk host name.\n'
             f'Takes place after "{CliLong.L2_REMOVE_DOMAIN}" and "{CliLong.L2_CASE}"'
    )
    parser.add_argument(
        CliLong.L2_REMOVE_DOMAIN,
        choices=L2RemoveDomain.list(),
        help=f'Handle the the domain name part of a neighbour name before matching it to a Checkmk host.\n'
             f'- {L2RemoveDomain.OFF} : dont touch the neighbour name, keep host name and domain part\n'
             f'- {L2RemoveDomain.ON}  : will remove the domain part from the neighbour name, keep only the host name part\n'
             f'- {L2RemoveDomain.AUTO} : try all of the above variants\n'
             f'Default: "{L2RemoveDomain.OFF}". Takes place after "{CliLong.L2_REMOVE_DOMAIN}" and before "{CliLong.L2_PREFIX}"',
    )
    parser.add_argument(
        CliLong.L2_SKIP_EXTERNAL, action='store_const', const=True,  # default=False,
        help='Skip L2 neighbours external to Checkmk (that have no matching host)',
    )
    parser.add_argument(
        CliLong.L3_DISPLAY_DEVICES, action='store_const', const=True,  # default=False,
        help='Use L3 device names as display name for interfaces in L3 topologies',
    )
    parser.add_argument(
        CliLong.L3_INCLUDE_HOSTS, action='store_const', const=True,  # default=False,
        help='Include hosts (single IP objects) in layer 3 topologies',
    )
    parser.add_argument(
        CliLong.L3_INCLUDE_LOOPBACK, action='store_const', const=True,  # default=False,
        help='Include loopback ip-addresses in layer 3 topologies',
    )
    parser.add_argument(
        CliLong.L3_SKIP_CIDR_0, action='store_const', const=True,  # default=False,
        help='Skip ip-address with CIDR "/0" in layer 3 topologies',
    )
    parser.add_argument(
        CliLong.L3_SKIP_CIDR_32_128, action='store_const', const=True,  # default=False,
        help='Skip ip-address with CIDR "/32" or "/128" in layer 3 topologies',
    )
    parser.add_argument(
        CliLong.L3_SKIP_IF, action='store_const', const=True,  # default=False,
        help='Dont show interface in layer 3 topologies',
    )
    parser.add_argument(
        CliLong.L3_SKIP_IP, action='store_const', const=True,  # default=False,
        help='Dont show ip-addresses in layer 3 topologies',
    )
    parser.add_argument(
        CliLong.L3_SKIP_PUBLIC, action='store_const', const=True,  # default=False,
        help='Skip public ip-addresses in layer 3 topologies',
    )
    parser.add_argument(
        CliLong.LOG_FILE, type=str,
        help='Set the log file. Default is "~/var/log/nvdct.log"',
    )
    parser.add_argument(
        CliLong.LOG_LEVEL,
        # nargs='+',
        choices=LogLevels.list(),
        # default='WARNING',
        help=f'Sets the log level. The default is "{LogLevels.WARNING}"'
    )
    parser.add_argument(
        CliLong.LOG_TO_STDOUT, action='store_const', const=True,  # default=False,
        help='Send log to stdout.',
    )
    parser.add_argument(
        CliLong.MIN_TOPOLOGY_AGE, type=int,
        help=f'The minimum number of days before a topology is deleted by "{CliLong.KEEP_MAX_TOPOLOGIES}"'
    )
    parser.add_argument(
        CliLong.PRE_FETCH, action='store_const', const=True,  # default=False,
        help=f'Try to fetch host data, with less API calls. Can improve {Backends.RESTAPI} backend\n'
             'performance',
    )
    parser.add_argument(
        CliLong.QUIET, action='store_const', const=True,  # default=False,
        help='Suppress all output to stdtout',
    )
    parser.add_argument(
        CliLong.TIME_FORMAT, type=str,
        help=f'Format string to render the time. (default: "{TIME_FORMAT_ARGPARSER}")',
    )
    parser.add_argument(
        CliLong.UPDATE_CONFIG, action='store_const', const=True,  # default=False,
        help='Adjusts options in config file.',
    )
    return parser.parse_args()
