#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2

# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-10-12
# File  : nvdct/lib/backends.py

# 2024-06-18: fixed host_exist returns always True if host was in host_cache, even with host=None
# 2024-09-25: fixed crash on missing "customer" section in site config file
# 2024-12-22: refactoring, leave only backend specific stuff in the backend
#             removed not strictly needed properties, renamed functions to better understand what the do
# 2025-01-21: added support for RESTAPI post requests (CMK >= 2.3.0p23, Werk #17003)
#             fixed REST API query for interface services
# 2025-01-25: fixed check if post can be used

from abc import abstractmethod
from ast import literal_eval
from collections.abc import Mapping, MutableMapping, MutableSet, MutableSequence, Sequence
from json import dumps
from pathlib import Path
from requests import session
from sys import exit as sys_exit
from typing import Dict, List, Tuple, Set

from livestatus import MultiSiteConnection, SiteConfigurations, SiteId

from lib.constants import (
    Backends,
    CACHE_INTERFACES_DATA,
    CacheItems,
    ExitCodes,
    HostFilter,
    IncludeExclude,
    InvPaths,
    L2Case,
    L2InvColumns,
    L2RemoveDomain,
    LiveStatusOperator,
    MIN_CMK_VERSION_POST,
    OMD_ROOT,
)
from lib.utils import (
    LOGGER,
    get_data_form_live_status,
    get_data_from_inventory,
    get_local_cmk_version,
)

HOST_EXIST: Dict = {'exists': True}


class HostCache:
    def __init__(
            self,
            backend: str,
            pre_fetch: bool,

    ):
        LOGGER.info(f'{backend} init HOST_CACHE')

        self.cache: Dict = {}
        self.neighbour_to_host: MutableMapping[str, str | None] = {}
        self._inventory_pre_fetch_list: List[str] = [InvPaths.INTERFACES]

        self.backend: str = str(backend)
        self.case: str = ''
        self.l2_neighbour_to_host_map: Dict[str, str] = {}
        self.l2_neighbour_replace_regex: List[Tuple[str, str]] = []
        self.pre_fetch: bool = bool(pre_fetch)
        self.prefix: str = ''
        self.remove_domain: str = L2RemoveDomain.OFF
        self.filter_include: MutableSequence[str] = []
        self.filter_exclude: MutableSequence[str] = []
        self.no_case_host_map: MutableMapping[str, str] = {}

        self.use_post = True

        if self.backend == f'[{Backends.RESTAPI}]' and get_local_cmk_version() < MIN_CMK_VERSION_POST:
            self.use_post = False
            LOGGER.info(
                f'{self.backend} Using get request, Checkmk version {get_local_cmk_version()} < {MIN_CMK_VERSION_POST}'
            )
        else:
            LOGGER.info(
                f'{self.backend} Using post request, Checkmk version {get_local_cmk_version()} >= {MIN_CMK_VERSION_POST}'
            )

        if self.pre_fetch:
            for host in self.query_all_hosts():
                self.cache[host] = HOST_EXIST.copy()

    def init_filter_lists(
            self,
            filter_by_folder: Mapping[str, set[str]],
            filter_by_host_label: Mapping[str, set[str]],
            filter_by_host_tag: Mapping[str, set[str]],
    ):
        for folder in filter_by_folder[IncludeExclude.INCLUDE]:
            self.filter_include += self.query_hosts_by_filter(HostFilter.FOLDER, folder, LiveStatusOperator.SUPERSET)

        for folder in filter_by_folder[IncludeExclude.EXCLUDE]:
            self.filter_exclude += self.query_hosts_by_filter(HostFilter.FOLDER, folder, LiveStatusOperator.SUPERSET)

        for host_label in filter_by_host_label[IncludeExclude.INCLUDE]:
            self.filter_include += self.query_hosts_by_filter(HostFilter.LABELS, host_label, LiveStatusOperator.EQUAL)

        for host_label in filter_by_host_label[IncludeExclude.EXCLUDE]:
            self.filter_exclude += self.query_hosts_by_filter(HostFilter.LABELS, host_label, LiveStatusOperator.EQUAL)

        for host_tag in filter_by_host_tag[IncludeExclude.INCLUDE]:
            self.filter_include += self.query_hosts_by_filter(HostFilter.TAGS, host_tag, LiveStatusOperator.EQUAL)

        for host_tag in filter_by_host_tag[IncludeExclude.EXCLUDE]:
            self.filter_exclude += self.query_hosts_by_filter(HostFilter.TAGS, host_tag, LiveStatusOperator.EQUAL)

        # drop any host from the include list that is on the exclude list
        # -> Issue: can render filter_include empty -> disabling include
        if self.filter_include:
            self.filter_include = [host for host in self.filter_include if host not in self.filter_exclude]
            if not self.filter_include:
                self.filter_include.append(self.filter_exclude[0])

        LOGGER.info(f'{self.backend} {IncludeExclude.INCLUDE} # of hosts: {len(self.filter_include)}')
        LOGGER.info(f'{self.backend} {IncludeExclude.EXCLUDE} # of hosts: {len(self.filter_exclude)}')
        LOGGER.debug(f'{self.backend} {IncludeExclude.INCLUDE} hosts: {self.filter_include}')
        LOGGER.debug(f'{self.backend} {IncludeExclude.EXCLUDE} hosts: {self.filter_exclude}')

    def init_neighbour_to_host_map(
            self,
            case: str,
            l2_host_map: Dict[str, str],
            prefix: str,
            remove_domain: str,
    ):
        self.case: str = case
        self.prefix: str = prefix
        self.remove_domain: str = remove_domain

        for host, data in self.cache.items():
            try:
                self.neighbour_to_host[data[CacheItems.inventory][InvPaths.LLDP_GLOBAL][L2InvColumns.GLOBALID]] = host
            except (KeyError, TypeError):
                pass
            try:
                self.neighbour_to_host[data[CacheItems.inventory][InvPaths.LLDP_GLOBAL][L2InvColumns.GLOBALNAME]] = host
            except (KeyError, TypeError):
                pass
            try:
                self.neighbour_to_host[data[CacheItems.inventory][InvPaths.CDP_GLOBAL][L2InvColumns.GLOBALNAME]] = host
            except (KeyError, TypeError):
                pass

        for neighbour, host in l2_host_map.items():
            self.neighbour_to_host[neighbour] = host

    def filter_host_list(self, host_list: MutableSequence[str] | MutableSet[str]) -> Sequence[str] | MutableSet[str]:
        if self.filter_include:
            host_list = [host for host in host_list if host in self.filter_include]
        if self.filter_exclude:
            host_list = [host for host in host_list if host not in self.filter_exclude]
        return host_list

    def get_inventory_data(self, hosts: List[str]) -> Dict[str, Dict]:
        """
        Returns a dictionary of hosts and there inventory data.
        Args:
            hosts: list of host names to return the inventory data for

        Returns:
            the inventory data as dictionary
        """

        # init inventory_data with None
        inventory_data: Dict[str, Dict | None] = {host: None for host in hosts}
        open_hosts = hosts.copy()
        while open_hosts:
            hosts_str, open_hosts = self.hosts_to_query(open_hosts)
            for host, inventory in self.query_inventory_data(hosts_str).items():
                inventory_data[host] = inventory
        return inventory_data

    def get_interface_data(self, hosts: List[str]) -> Dict[str, Dict | None]:
        """
        Returns Dictionary of hosts and there interface services from CMK.
        The interface information consists of the "Item", the "Description (summary)" and the service details
        Args:
            hosts:  lit of host names to return the interface data for

        Returns:
            dictionary of the interface data with the host -> item as key
        """
        host_data: Dict[str, any] = {}  # to make pylint happy
        # init host_data with None
        for host in hosts:
            host_data[host] = None
        open_hosts = hosts.copy()
        while open_hosts:
            hosts_str, open_hosts = self.hosts_to_query(open_hosts)
            host_data.update(self.query_interface_data(hosts_str))

        return host_data

    def host_exists(self, host: str) -> bool:
        """
        Returns True if host exists in CMK, else False
        """
        try:
            return bool(self.cache[host])
        except KeyError:
            pass

        if self.pre_fetch:
            return False

        # get host from CMK and init host in cache
        if exists := self.query_host(host):
            self.cache[host] = HOST_EXIST.copy()
        else:
            self.cache[host] = None

        return exists

    def is_host_allowed(self, host: str) -> bool:
        if self.filter_include and host not in self.filter_include:
            return False
        if self.filter_exclude and host in self.filter_exclude:
            return False

        return True

    def fill_cache(self, hosts: List[str]) -> None:
        """
        Gets the host data from CMK and puts them in the host cache. Data collected:
        - inventory
        - interfaces

        Args:
            hosts: List of hosts the get data from CMK

        Returns: None, the data is directly writen to self.cache
        """

        inventory_of_hosts: Mapping[str, Dict | None] = self.get_inventory_data(hosts=hosts)
        if inventory_of_hosts:
            for host, inventory in inventory_of_hosts.items():
                if host not in self.cache:
                    self.cache[host] = HOST_EXIST.copy()
                self.cache[host][CacheItems.inventory] = {}
                self.cache[host][CacheItems.inventory].update({
                    entry: get_data_from_inventory(
                        inventory=inventory,
                        raw_path=entry
                    ) for entry in self._inventory_pre_fetch_list
                })

        interfaces_of_hosts: Mapping[str, Mapping | None] = self.get_interface_data(hosts)
        for host, interfaces in interfaces_of_hosts.items():
            if host not in self.cache:
                self.cache[host] = HOST_EXIST.copy()
            if not self.cache[host].get(CacheItems.interfaces):
                self.cache[host][CacheItems.interfaces] = {}
            self.cache[host][CacheItems.interfaces][CACHE_INTERFACES_DATA] = interfaces

    def get_data(self, host: str, item: CacheItems, path: str) -> Mapping | Sequence | None:
        """
        Returns data from self.cache. If the cache for "host" is empty, data will be fetched from CMK
        Args:
            host: host to get data from cache
            item: item in cache (inventory/interface)
            path: path in cache to data

        Returns:
            the requested data or None
        """
        if self.host_exists(host=host):
            if self.cache[host] == HOST_EXIST:
                LOGGER.info(f'{self.backend} fetch data for: {host}')
                self.fill_cache(hosts=[host])
            try:
                return self.cache[host][item][path]
            except (KeyError, TypeError):
                return None

        return None

    def add_inventory_path(self, path: str) -> None:
        self._inventory_pre_fetch_list = list(set(self._inventory_pre_fetch_list + [path]))

    def get_host_from_neighbour(
            self,
            raw_neighbour: str,
            neighbour: str,
            neighbour_id: str | None = None,
    ) -> str | None:
        """
        Tries to get the CMK host name from a L2 neighbour name. It will test:
        - the neighbour id
        - the raw neighbour name
        - the adjusted neighbour name
        - map the neighbour to a host via L2_NEIGHBOUR_TO_HOST_MAP
        - the neighbour without domain name
        - the neighbour in UPPER case (without domain)
        - the neighbour in lower case (including domain)
        - the neighbour with prefix (UPPER/lower case, with/without domain)
        Args:
            raw_neighbour: the L2 neighbour name "as is" in the HW/SW inventory
            neighbour: the L2 neighbour name to find a CMK host for
            neighbour_id: the L2 neighbour id to find a CMK host for (LLDP only)

        Returns:
            The CMK host name for the L2 neighbour or None if no host is found
        """

        try:
            match = self.neighbour_to_host[neighbour_id]
            LOGGER.debug(f'{self.backend} Match by neighbour id: |{neighbour_id}| -> |{match}|')
            return match
        except KeyError:
            pass

        try:
            match = self.neighbour_to_host[raw_neighbour]
            LOGGER.debug(f'{self.backend} Match by raw neighbour: |{raw_neighbour}| -> |{match}|')
            return match
        except KeyError:
            pass

        try:
            match = self.neighbour_to_host[neighbour]
            LOGGER.debug(f'{self.backend} Match by neighbour: |{neighbour}| -> |{match}|')
            return match
        except KeyError:
            pass

        LOGGER.debug(f'{self.backend} neighbour not in list: |{neighbour}|')

        host: str = neighbour

        possible_hosts: Set[str] | List[str] = set()

        host_no_domain: str = host.split('.')[0]
        match self.remove_domain:
            case L2RemoveDomain.ON:
                LOGGER.debug(f'{self.backend} Remove domain: {host} -> {host_no_domain}')
                host = host_no_domain
                possible_hosts.add(host_no_domain)
            case L2RemoveDomain.AUTO:
                possible_hosts.add(host)
                possible_hosts.add(host_no_domain)
            case L2RemoveDomain.OFF | _:
                possible_hosts.add(host)

        match self.case:
            case L2Case.UPPER:
                if not '.' in host:  # use UPPER only for names without domain
                    LOGGER.debug(f'{self.backend} Change neighbour to upper case: {host} -> {host.upper()}')
                    possible_hosts.add(host.upper)
            case L2Case.LOWER:
                LOGGER.debug(f'{self.backend} Change neighbour to lower case: {host} -> {host.lower()}')
                possible_hosts.add(host.lower())
            case L2Case.AUTO:
                possible_hosts.add(host)
                possible_hosts.add(host.lower())
                possible_hosts.add(host_no_domain.lower())
                possible_hosts.add(host_no_domain.upper())
            case L2Case.OFF | _:
                possible_hosts.add(host)

        possible_hosts = [f'{self.prefix}{host}' for host in possible_hosts]
        possible_hosts = set(self.filter_host_list(possible_hosts))
        # try longest match first
        possible_hosts = list(possible_hosts)
        possible_hosts.sort(key=len, reverse=True)

        for entry in possible_hosts:
            if self.host_exists(entry):
                self.neighbour_to_host[neighbour] = entry
                LOGGER.debug(f'{self.backend} Matched neighbour to host: |{neighbour}| -> |{entry}|')
                return entry

        if self.case in [L2Case.AUTO, L2Case.INSENSITIVE]:
            if not self.no_case_host_map:
                self.no_case_host_map = {host.lower():host for host in self.cache}
            for entry in possible_hosts:
                if entry.lower() in self.no_case_host_map:
                    self.neighbour_to_host[neighbour] = self.no_case_host_map[entry.lower()]
                    LOGGER.debug(
                        f'{self.backend} Matched neighbour to host: '
                        f'|{neighbour}| -> |{self.no_case_host_map[entry.lower()]}|'
                    )
                    return entry

        self.neighbour_to_host[neighbour] = None
        LOGGER.debug(f'{self.backend} No match found for neighbour: |{neighbour}|')
        return None

    def hosts_to_query(self, hosts: List[str]) -> Tuple[str, List[str]]:
        if self.use_post:
            return '|'.join(hosts), []

        # WORKAROUND for: Apache HTTP Error 414: Request URI too long
        # https://httpd.apache.org/docs/current/mod/core.html#limitrequestfieldsize
        # Default:	LimitRequestFieldSize 8190
        # max seems to be 8013 (inventory)/7831 (interfaces), so 7800 is a save distance from it
        _max_str_len = 7800

        # 8190 - 8013 - 168 = 9 chrs  -> Inventory
        # 8190 - 7813 - 356 = 21 chrs -> Interfaces
        # host (19/19): http://localhost:80 -> sould not count
        # uri (57/60): /build/check_mk/api/1.0/domain-types/host/collections/all / /build/check_mk/api/1.0/domain-types/service/collections/all  # noqa: E501
        # query (77/231): ?query=%7B%22op%22%3A+%22~~%22%2C+%22left%22%3A+%22name%22%2C+%22right%22%3A+ / ?query=%7B%22op%22%3A+%22and%22%2C+%22expr%22%3A+%5B%7B%22op%22%3A+%22~%22%2C+%22left%22%3A+%22description%22%2C+%22right%22%3A+%22Interface+%22%7D%2C%7B%22op%22%3A+%22~~%22%2C+%22left%22%3A+%22host_name%22%2C+%22right%22%3A+%22%5E  # noqa: E501
        # collumns (34/65): &columns=name&columns=mk_inventory / &columns=host_name&columns=description&columns=long_plugin_output  # noqa: E501
        # %22 -> "
        # %24 -> $
        # %2C -> ,
        # %3A -> :
        # %5B -> [
        # %5D -> ]
        # %7B -> {
        # %7D -> }

        hosts = [host for host in hosts if self.is_host_allowed(host)]
        open_hosts = []
        temp_hosts: Sequence[str] = [f'^{host}$' for host in hosts]
        hosts_str = '|'.join(temp_hosts)
        # 6 comes from 3 chrs ($|^) between hosts, coded as 9 chr(%24%7C%5E) in the URL
        # 3 are already in the str, so we need to add 6 for each host
        if len(hosts_str) > _max_str_len - len(hosts) * 6:
            open_hosts = hosts.copy()
            hosts_str = f'^{hosts.pop()}$'
            count = 1
            for host in hosts:
                count += 1
                if len(hosts_str) + 9 + len(host) + (6 * count) < _max_str_len:
                    hosts_str = hosts_str + f'|^{host}$'
                    open_hosts.remove(host)
                else:
                    break

        LOGGER.debug(f'{self.backend} hosts len: {len(hosts_str) + (len(hosts) - len(open_hosts)) * 6}')
        LOGGER.debug(f'{self.backend} open hosts {open_hosts}')

        return hosts_str, open_hosts

    @abstractmethod
    def query_host(self, host: str) -> bool:
        """
        Query Livestatus for "host"

        Args:
             host: CMK host name to query livestus for
        Returns:
            True: if host was found
            False: if host is not found
        """
        raise NotImplementedError

    @abstractmethod
    def query_all_hosts(self) -> Sequence[str]:
        """
        Queries Livestatus for a list of all hosts
        Returns:
            List of all hosts
        """
        raise NotImplementedError

    @abstractmethod
    def query_hosts_by_filter(self, filter_name: str, filter_value: str, filter_operator: str) -> MutableSequence[str]:
        """
        Queries Livestatus for a list of hosts filtered by a host label
        Args:
            filter_name: The column name to filter on
            filter_value: the value for the attribute to filter on
            filter_operator: Live status filter operator
        Returns: List of hosts
        """
        raise NotImplementedError

    @abstractmethod
    def query_inventory_data(self, hosts: str) -> Dict[str, Dict]:
        raise NotImplementedError

    @abstractmethod
    def query_interface_data(self, hosts: str) -> Dict[str, Dict]:
        raise NotImplementedError


class HostCacheLiveStatus(HostCache):
    def __init__(
            self,
            pre_fetch: bool,
            backend: str = f'[{Backends.LIVESTATUS}]',
    ):
        self.backend = backend
        super().__init__(
            pre_fetch=pre_fetch,
            backend=self.backend,
        )

    def get_raw_data(self, query: str) -> any:
        return get_data_form_live_status(query=query)

    def query_host(self, host: str) -> bool:
        query = (
            'GET hosts\n'
            'Columns: name\n'
            'OutputFormat: python3\n'
            f'Filter: name = {host}\n'
        )
        data: Sequence[Sequence[str]] = self.get_raw_data(query=query)
        LOGGER.debug(f'{self.backend} data for host {host}: {data}')
        if [host] in data:
            LOGGER.debug(f'{self.backend} Host found in Checkmk: {host} ')
            return True
        LOGGER.warning(f'{self.backend} Host not found in Checkmk: {host}')
        return False

    def query_all_hosts(self) -> Sequence[str]:
        query = (
            'GET hosts\n'
            'Columns: name\n'
            'OutputFormat: python3\n'
        )
        data: Sequence[Sequence[str]] = self.get_raw_data(query=query)
        if data:
            LOGGER.info(f'{self.backend} # of hosts found: {len(data)}')
            return [host[0] for host in data]

        LOGGER.warning(f'{self.backend} no hosts found')
        return []

    def query_hosts_by_filter(self, filter_name: str, filter_value: str, filter_operator: str) -> MutableSequence[str]:
        query = (
            'GET hosts\n'
            'Columns: name\n'
            'OutputFormat: python3\n'
            f'Filter: {filter_name} {filter_operator} {filter_value}\n'
        )
        data: Sequence[Sequence[str]] = self.get_raw_data(query=query)
        LOGGER.debug(f'{self.backend} hosts matching filter {filter_value}: {data}')
        if data:
            LOGGER.info(f'{self.backend} # of hosts matching filter {filter_value}: {len(data)}')
            return [host[0] for host in data]

        LOGGER.warning(f'{self.backend} no hosts found matching filter: {filter_value}')
        return []

    def query_inventory_data(self, hosts: str) -> Dict[str, Dict]:
        query = (
            'GET hosts\n'
            'Columns: name mk_inventory\n'
            'OutputFormat: python3\n'
            f'Filter: name ~~ {hosts}\n'
        )
        inventory_data = {}
        data: Sequence[Tuple[str, bytes]] = self.get_raw_data(query=query)
        LOGGER.debug(f'{self.backend} data for hosts {hosts}: {data}')
        if data:
            for host, inventory in data:
                if not inventory:
                    LOGGER.warning(f'{self.backend} No inventory data found for host: {host}')
                    continue
                inventory = literal_eval(inventory.decode('utf-8'))
                inventory_data[host] = inventory
        else:
            LOGGER.warning(f'{self.backend} Device: {hosts}: no inventory data found!')

        return inventory_data

    def query_interface_data(self, hosts: str) -> Dict[str, Dict]:
        query = (
            'GET services\n'
            'Columns: host_name description long_plugin_output\n'
            'Filter: description ~ ^Interface\n'
            f'Filter: host_name ~~ {hosts}\n'
            'OutputFormat: python3\n'
        )
        interface_data = {}
        data: List[Tuple[str, str, str]] = self.get_raw_data(query=query)
        LOGGER.debug(f'{self.backend} interface data for hosts: {hosts}: {data}')
        if data:
            for host, description, long_plugin_output in data:
                if interface_data.get(host) is None:
                    interface_data[host] = {}
                interface_data[host][description[10:]] = {  # remove 'Interface ' from description
                    'long_plugin_output': long_plugin_output.split('\\n')
                }
        else:
            LOGGER.warning(f'{self.backend} No Interfaces items found for hosts: {hosts}')

        return interface_data


class HostCacheMultiSite(HostCacheLiveStatus):
    def __init__(
            self,
            pre_fetch: bool,
            filter_sites: str | None = None,
            sites: List[str] | None = None,
            filter_customers: str | None = None,
            customers: List[str] = None,
    ):
        if not sites:
            sites = []
        self.backend = f'[{Backends.MULTISITE}]'
        self.sites: SiteConfigurations = SiteConfigurations({})
        self.get_sites()
        self.filter_sites(filter_sites, sites)
        self.filter_costumers(filter_customers, customers)
        LOGGER.info(f'{self.backend} Create livestatus connection(s)')
        self.c = MultiSiteConnection(self.sites)
        # self.c.set_prepend_site(False)  # is default
        # self.c.parallelize = True   # is default
        self.dead_sites = [site['site']['alias'] for site in self.c.dead_sites().values()]
        if self.dead_sites:
            dead_sites = ', '.join(self.dead_sites)
            LOGGER.warning(f'{self.backend} use of dead site(s) is disabled: {dead_sites}')
            self.c.set_only_sites(self.c.alive_sites())
        super().__init__(
            pre_fetch=pre_fetch,
            backend=self.backend,
        )

    def get_raw_data(self, query: str) -> object:
        return self.c.query(query=query)

    # https://github.com/Checkmk/checkmk/blob/master/packages/cmk-livestatus-client/example_multisite.py
    def get_sites(self):
        sites_mk = Path(f'{OMD_ROOT}/etc/check_mk/multisite.d/sites.mk')
        socket_path = f'unix:{OMD_ROOT}/tmp/run/live'
        if sites_mk.exists():
            # make eval() "secure"
            # https://realpython.com/python-eval-function/#minimizing-the-security-issues-of-eval
            _code = compile(sites_mk.read_text(), "<string>", "eval")
            allowed_names = ['sites', 'update']
            for name in _code.co_names:
                if name not in allowed_names:
                    raise NameError(f'Use of {name} in {sites_mk.name} not allowed.')

            sites_raw: Dict = {}
            eval(sites_mk.read_text(), {"__builtins__": {}}, {"sites": sites_raw})

            for site, data in sites_raw.items():
                self.sites.update({site: {
                    'alias': data['alias'],
                    'timeout': data['timeout'],
                    'customer': data.get('customer')  # needed to filter by customer
                    # 'nagios_url': '/nagios/',
                }})
                if data['socket'] == ('local', None):
                    self.sites[site]['socket'] = socket_path
                else:
                    protocol, socket = data['socket']
                    address, port = socket['address']
                    self.sites[site]['socket'] = f'{protocol}:{address}:{port}'
                    self.sites[site]['tls'] = socket['tls']
                LOGGER.info(
                    f'{self.backend} Site {site} found, '
                    f'Socket: {self.sites[site]["socket"]}, '
                    f'TLS: {self.sites[site].get("tls", "N/A")}.'
                )
        else:
            self.sites.update({SiteId('local'): {
                'alias': 'Local site',
                'timeout': 5,
                # 'nagios_url': '/nagios/',
                'socket': socket_path
            }})

            LOGGER.critical(
                f'{self.backend} file {str(sites_mk.absolute())} not found. Falling back to '
                'local site only. Try -b RESTAPI if you have a distributed environment.'
            )

    def filter_sites(self, filter_: str | None, sites: List[str]):
        match filter_:
            case IncludeExclude.INCLUDE:
                self.sites = {site: data for site, data in self.sites.items() if site in sites}
            case IncludeExclude.EXCLUDE:
                self.sites = {site: data for site, data in self.sites.items() if site not in sites}
            case _:
                return

    def filter_costumers(self, filter_: str | None, costumers: List[str]):
        match filter_:
            case IncludeExclude.INCLUDE:
                self.sites = {
                    site: data for site, data in self.sites.items() if data.get('customer') in costumers
                }
            case IncludeExclude.EXCLUDE:
                self.sites = {
                    site: data for site, data in self.sites.items() if data.get('customer') not in costumers
                }
            case _:
                return


class HostCacheRestApi(HostCache):
    def __init__(
            self,
            pre_fetch: bool,
            api_port: int,
            filter_sites: str | None = None,
            sites: List[str] | None = None,
    ):
        if not sites:
            sites = []
        self.backend = f'[{Backends.RESTAPI}]'
        LOGGER.debug(f'{self.backend} init backend')

        try:
            self.__secret = Path(
                f'{OMD_ROOT}/var/check_mk/web/automation/automation.secret'
            ).read_text().strip('\n')
        except FileNotFoundError as e:
            LOGGER.exception(f'{self.backend} automation.secret not found, {e}')
            print(f'{self.backend} automation secret not found, {e}')
            sys_exit(ExitCodes.AUTOMATION_SECRET_NOT_FOUND)

        self.__api_port = api_port
        self.__hostname = 'localhost'
        self.__site = OMD_ROOT.split('/')[-1]
        self.__api_url = f"http://{self.__hostname}:{self.__api_port}/{self.__site}/check_mk/api/1.0"
        self.__user = 'automation'

        LOGGER.info(f'{self.backend} Create REST API session')
        self.__session = session()
        self.__session.headers['Authorization'] = f"Bearer {self.__user} {self.__secret}"
        self.__session.headers['Accept'] = 'application/json'
        self.__session.headers['Content-Type'] = 'application/json'
        self.sites: MutableSequence[str] = self.query_sites()
        self.filter_sites(filter_=filter_sites, sites=sites)
        LOGGER.info(f'{self.backend} filtered sites: {self.sites}')
        super().__init__(
            pre_fetch=pre_fetch,
            backend=self.backend,
        )

    def get_raw_data(self, url: str, params: Mapping[str, object] | None, allow_post: bool = False):
        if allow_post and self.use_post:
            resp = self.__session.post(
                url=url,
                data=dumps(params),
                # timeout=3,
            )
        else:
            resp = self.__session.get(
                url=url,
                params=params,
                # timeout=3,
            )
        LOGGER.debug(f'{self.backend} raw data: {resp.text}')

        if resp.status_code == 200:
            return resp.json()
        else:
            LOGGER.warning(f'{self.backend} response: {resp.status_code}')

    def query_sites(self) -> MutableSequence[str]:
        LOGGER.debug(f'{self.backend} get_sites')
        url = f"{self.__api_url}/domain-types/site_connection/collections/all"
        sites = []
        if raw_data := self.get_raw_data(url, None):
            raw_sites = raw_data.get("value")
            sites = [site.get('id') for site in raw_sites]
            LOGGER.debug(f'{self.backend} sites: {sites}')
        else:
            LOGGER.warning(f'{self.backend} got no site information!')

        return sites

    def filter_sites(self, filter_: str | None, sites: List[str]):
        match filter_:
            case IncludeExclude.INCLUDE:
                self.sites = [site for site in self.sites if site in sites]
            case IncludeExclude.EXCLUDE:
                self.sites = [site for site in self.sites if site not in sites]
            case _:
                return

    def query_host(self, host: str) -> bool:
        query = '{"op": "=", "left": "name", "right": "' + host + '"}'
        url = f'{self.__api_url}/domain-types/host/collections/all'
        params = {
            'query': query,
            'columns': ['name'],
            'sites': self.sites,
        }

        if raw_data := self.get_raw_data(url, params, self.use_post):
            try:
                data = raw_data['value'][0]['extensions']['name']
                LOGGER.debug(f'{self.backend} data for host {host}: {data}')
            except IndexError:
                LOGGER.warning(f'{self.backend} Host not found in Checkmk: {host}')
            else:
                if data == host:
                    LOGGER.debug(f'{self.backend} Host found in Checkmk: {host}')
                    return True

        return False

    def query_all_hosts(self) -> Sequence[str]:
        url = f'{self.__api_url}/domain-types/host/collections/all'
        params = {
            'columns': ['name'],
            'sites': self.sites,
        }

        if raw_data := self.get_raw_data(url, params, self.use_post):
            if data := raw_data.get('value', []):
                LOGGER.info(f'{self.backend} # of hosts found: {len(data)}')
                return [host.get('extensions', {}).get('name') for host in data]

        LOGGER.warning(f'{self.backend} no hosts found')
        return []

    def query_hosts_by_filter(self, filter_name: str, filter_value: str, filter_operator: str) -> MutableSequence[str]:
        query = '{"op": "' + filter_operator + '", "left": "' + filter_name + '", "right": "' + filter_value + '"}'
        url = f'{self.__api_url}/domain-types/host/collections/all'
        params = {
            'columns': ['name'],
            'query': query,
            'sites': self.sites,
        }

        raw_data = self.get_raw_data(url, params, self.use_post)
        LOGGER.debug(f'{self.backend} hosts matching filter {filter_value}: {raw_data}')
        if raw_data:
            if data := raw_data.get('value'):
                LOGGER.info(f'{self.backend} # of hosts matching filter {filter_value}: {len(data)}')
                return [host['extensions']['name'] for host in data]

        LOGGER.warning(f'{self.backend} no hosts found matching filter: {filter_value}')
        return []

    def query_inventory_data(self, hosts: str) -> Dict[str, Dict]:
        query = '{"op": "~~", "left": "name", "right": "' + hosts + '"}'
        url = f"{self.__api_url}/domain-types/host/collections/all"
        params = {
            'query': query,
            'columns': ['name', 'mk_inventory'],
            'sites': self.sites,
        }

        inventory_data = {}

        raw_data = self.get_raw_data(url, params, self.use_post)
        LOGGER.debug(f'{self.backend} raw inventory data: {raw_data}')
        if raw_data:
            if data := raw_data.get('value', []):
                for raw_host in data:
                    if host := raw_host.get('extensions', {}).get('name'):
                        inventory = raw_host['extensions'].get('mk_inventory')
                        if not inventory:
                            LOGGER.warning(f'{self.backend} No inventory data found for host: {host}!')
                        inventory_data[host] = inventory

        return inventory_data

    def query_interface_data(self, hosts: str) -> Dict[str, Dict]:
        query_host = '{"op": "~~", "left": "host_name", "right": "' + hosts + '"}'
        query_item = '{"op": "~", "left": "description", "right": "^Interface "}'
        query = '{"op": "and", "expr": [' + query_host + ',' + query_item +']}'
        url = f'{self.__api_url}/domain-types/service/collections/all'
        params = {
            'query': query,
            'columns': ['host_name', 'description', 'long_plugin_output'],
            'sites': self.sites,
        }

        interface_data = {}

        raw_data = self.get_raw_data(url, params) #  , self.use_post  -> Endpoint not changed to Post :-(
        LOGGER.debug(f'{self.backend} interface data for host(s) {hosts}: {raw_data}')
        if raw_data:
            if data := raw_data.get('value', []):
                for raw_service in data:
                    LOGGER.debug(f'{self.backend} data for service: {raw_service}')
                    service = raw_service.get('extensions')
                    host, description, long_plugin_output = service.values()
                    if interface_data.get(host) is None:
                        # LOGGER.warning(f'{self.backend} No Interfaces items found for hosts {host}')
                        interface_data[host] = {}
                    interface_data[host][description[10:]] = {
                        'long_plugin_output': long_plugin_output.split('\\n')
                    }

        return interface_data
