#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-10-12
# File  : nvdct/lib/settings.py

# fixed path to default user data file
# 2024-12-17: fixed wrong import for OMD_ROOT (copy&paste) (ThX to BH2005@forum.checkmk.com)
# 2025-01-19: moved --check-toml to utils/get_data_from_toml

from collections.abc import Mapping
from ipaddress import AddressValueError, NetmaskValueError, ip_address, ip_network
from logging import CRITICAL, DEBUG, ERROR, FATAL, INFO, WARNING
from pathlib import Path
from time import strftime
from typing import Dict, List, NamedTuple, Set, Tuple

from lib.constants import (
    API_PORT_DEFAULT,
    Backends,
    CONFIG_FILE,
    CONFIG_PATH_DEFAULT,
    EmblemNames,
    EmblemValues,
    IncludeExclude,
    L2Case,
    L2IgnoreMismatch,
    L2RemoveDomain,
    Layers,
    LOGGER,
    LOG_FILE_DEFAULT,
    LOG_PATH_DEFAULT,
    LogLevels,
    TIME_FORMAT_DEFAULT,
    TomlSections,
    TomlSettings,
)
from lib.utils import (
    get_data_from_toml,
    get_local_cmk_api_port,
    is_valid_customer_name,
    is_valid_hostname,
    is_valid_output_directory,
    is_valid_site_name,
)


class Emblems(NamedTuple):
    host_node: str
    ip_address: str
    ip_network: str
    l3_replace: str
    l3_summarize: str
    service_node: str


class Thickness(NamedTuple):
    speed: int
    thickness: int


class StaticConnection(NamedTuple):
    left_host: str
    left_service: str
    right_host: str
    right_service: str


class Wildcard(NamedTuple):
    bit_pattern: int
    int_ip_address: int
    int_wildcard: int
    ip_address: str
    wildcard: str


class Settings:
    def __init__(
            self,
            cli_args: Mapping[str, object],
    ):
        # cli defaults
        self.__settings = {
            TomlSettings.API_PORT: None,
            TomlSettings.BACKEND: Backends.MULTISITE,
            TomlSettings.CHECK_CONFIG: False,
            TomlSettings.CONFIG: f'{CONFIG_PATH_DEFAULT}/{CONFIG_FILE}',
            TomlSettings.DEFAULT: False,
            TomlSettings.DONT_COMPARE: False,
            TomlSettings.FILTER_CUSTOMERS: None,
            TomlSettings.FILTER_SITES: None,
            TomlSettings.KEEP_MAX_TOPOLOGIES: False,
            TomlSettings.L2_CASE: None,
            TomlSettings.L2_DISPLAY_PORTS: False,
            TomlSettings.L2_IGNORE_MISMATCH: [],
            TomlSettings.L2_DISPLAY_NEIGHBOURS: False,
            TomlSettings.L2_PREFIX: None,
            TomlSettings.L2_REMOVE_DOMAIN: None,
            TomlSettings.L2_SKIP_EXTERNAL: False,
            TomlSettings.L3_DISPLAY_DEVICES: False,
            TomlSettings.L3_INCLUDE_HOSTS: False,
            TomlSettings.L3_INCLUDE_LOOPBACK: False,
            TomlSettings.L3_SKIP_CIDR_0: False,
            TomlSettings.L3_SKIP_CIDR_32_128: False,
            TomlSettings.L3_SKIP_IF: False,
            TomlSettings.L3_SKIP_IP: False,
            TomlSettings.L3_SKIP_PUBLIC: False,
            TomlSettings.LAYERS: [],
            TomlSettings.LOG_FILE: LOG_FILE_DEFAULT,
            TomlSettings.LOG_LEVEL: LogLevels.WARNING,
            TomlSettings.LOG_TO_STDOUT: False,
            TomlSettings.MIN_TOPOLOGY_AGE: 0,
            TomlSettings.OUTPUT_DIRECTORY: None,
            TomlSettings.PRE_FETCH: False,
            TomlSettings.QUIET: False,
            TomlSettings.TIME_FORMAT: TIME_FORMAT_DEFAULT,
            TomlSettings.UPDATE_CONFIG: False,
        }
        # args in the form {'s, __seed_devices': 'CORE01', 'p, __path_in_inventory': None, ... }}
        # we will remove 's, __'
        self.__args = (
            {k.split(',')[-1].strip(' ').strip('_'): v for k, v in cli_args.items() if v}
        )

        self.__user_data = get_data_from_toml(
            file=self.__args.get(TomlSettings.CONFIG, self.config),
            check_only=bool(self.__args.get(TomlSettings.CHECK_CONFIG)),
        )

        # defaults -> overridden by toml -> overridden by cli
        self.__settings.update(self.__user_data.get(TomlSections.SETTINGS, {}))
        self.__settings.update(self.__args)

        self.__api_port: int | None = None

        # init user data with defaults
        self.__emblems: Emblems | None = None
        self.__filter_by_customer: List[str] | None = None
        self.__filter_by_folder: Dict[str, Set[str]] | None = None
        self.__filter_by_host_label: Dict[str, Set[str]] | None = None
        self.__filter_by_host_tag: Dict[str, Set[str]] | None = None
        self.__filter_by_site: List[str] | None = None
        self.__l2_drop_neighbours: List[str] | None = None
        self.__l2_ignore_mismatch: List[str] | None = None
        self.__l2_neighbour_replace_regex: List[Tuple[str, str]] | None = None
        self.__l2_neighbour_to_host_map: Dict[str, str] | None = None
        self.__l2_seed_devices: List[str] | None = None
        self.__l3_ignore_hosts: List[str] | None = None
        self.__l3_ignore_ip: List[ip_network] | None = None
        self.__l3_replace: Dict[str, str] | None = None
        self.__l3_summarize: List[ip_network] | None = None
        self.__l3v4_ignore_wildcard: List[Wildcard] | None = None
        self.__layers: List[str] | None = None
        self.__map_speed_to_thickness: List[Thickness] | None = None
        self.__protected_topologies: List[str] | None = None
        self.__static_connections: List[StaticConnection] | None = None

    #
    # CLI settings
    #
    @property  # --api-port
    def api_port(self) -> int:
        if self.__api_port is None:
            if self.__settings.get(TomlSettings.API_PORT):
                self.__api_port = int(self.__settings.get(TomlSettings.API_PORT))
            else:
                self.__api_port = get_local_cmk_api_port()
            if self.__api_port is None:
               self.__api_port = API_PORT_DEFAULT

        return self.__api_port

    @property  # -b --backend
    def backend(self) -> str:
        if str(self.__settings[TomlSettings.BACKEND]) in Backends.list():
            return str(self.__settings[TomlSettings.BACKEND])
        else:  # fallback to defaukt -> exit ??
            LOGGER.error(
                f'Unknown backend: {self.__settings[TomlSettings.BACKEND]}. Accepted backends are: '
                f'{Backends.LIVESTATUS}, {Backends.MULTISITE}, {Backends.RESTAPI}. Falling back to {Backends.MULTISITE}.'
            )
            return Backends.MULTISITE

    @property  # --check-config
    def check_config(self) -> bool:
        return bool(self.__settings[TomlSettings.CHECK_CONFIG])

    @property  # --config
    def config(self) -> str:
        return str(self.__settings[TomlSettings.CONFIG])

    @property  # -d --default
    def default(self) -> bool:
        return bool(self.__settings[TomlSettings.DEFAULT])

    @property  # --dont-compare
    def dont_compare(self) -> bool:
        return bool(self.__settings[TomlSettings.DONT_COMPARE])

    @property  # --filter-customers
    def filter_customers(self) -> str | None:
        if self.__settings[TomlSettings.FILTER_CUSTOMERS] in IncludeExclude.list():
            return self.__settings[TomlSettings.FILTER_CUSTOMERS]
        elif self.__settings[TomlSettings.FILTER_CUSTOMERS] is not None:
            LOGGER.error(
                f'Wrong setting for "{TomlSettings.FILTER_CUSTOMERS}": '
                f'{self.__settings[TomlSettings.FILTER_CUSTOMERS]}, supported settings {IncludeExclude.INCLUDE}|{IncludeExclude.EXCLUDE}.'
            )
        return None

    @property  # --filter-sites
    def filter_sites(self) -> str | None:
        if self.__settings[TomlSettings.FILTER_SITES] in IncludeExclude.list():
            return self.__settings[TomlSettings.FILTER_SITES]
        elif self.__settings[TomlSettings.FILTER_SITES] is not None:
            LOGGER.error(
                f'Wrong setting for "{TomlSettings.FILTER_SITES}": '
                f'{self.__settings[TomlSettings.FILTER_SITES]}, supported settings {IncludeExclude.INCLUDE}|{IncludeExclude.EXCLUDE}.'
            )
        return None

    @property  # --l2-case
    def l2_case(self) -> str | None:
        if self.__settings[TomlSettings.L2_CASE] in L2Case.list():
            return self.__settings[TomlSettings.L2_CASE]
        elif self.__settings[TomlSettings.L2_CASE] is not None:
            LOGGER.error(
                    f'Unknown case setting {self.__settings[TomlSettings.L2_CASE]}. '
                    f'Accepted are {L2Case.LOWER}|{L2Case.UPPER}|{L2Case.INSENSITIVE}|{L2Case.AUTO}|{L2Case.OFF}. Falling back to "OFF" (no change).'
                )
        return None

    @property  # --l2-display-ports
    def l2_display_ports(self) -> bool:
        return bool(self.__settings[TomlSettings.L2_DISPLAY_PORTS])

    @property  # --l2-display-neighbours
    def l2_display_neighbours(self) -> bool:
        return bool(self.__settings[TomlSettings.L2_DISPLAY_NEIGHBOURS])

    @property  # --l2-ignore-mismatch
    def l2_ignore_mismatch(self) -> List[str]:
        if self.__l2_ignore_mismatch is None:
            self.__l2_ignore_mismatch = list(set(self.__settings[TomlSettings.L2_IGNORE_MISMATCH]))
            self.__l2_ignore_mismatch = [entry for entry in self.__l2_ignore_mismatch if entry in L2IgnoreMismatch.list()]
            if len(self.__layers) != len(self.__settings[TomlSettings.LAYERS]):
                LOGGER.error(f'Wrong/duplicate l2_mismatch_ignore options ignored, {self.__settings[TomlSettings.L2_IGNORE_MISMATCH]}')
        return self.__l2_ignore_mismatch

    @property  # --l2-prefix
    def l2_prefix(self) -> str:
        if self.__settings[TomlSettings.L2_PREFIX] is not None:
            return str(self.__settings[TomlSettings.L2_PREFIX])
        return ''

    @property  # --l2-remove-domain
    def l2_remove_domain(self) -> str:
        if self.__settings[TomlSettings.L2_REMOVE_DOMAIN] in L2RemoveDomain.list():
            return self.__settings[TomlSettings.L2_REMOVE_DOMAIN]
        else:
            self.__settings[TomlSettings.L2_REMOVE_DOMAIN] = L2RemoveDomain.OFF
            return L2RemoveDomain.OFF

    @property  # --l2-skip-external
    def l2_skip_external(self) -> bool:
        return bool(self.__settings[TomlSettings.L2_SKIP_EXTERNAL])

    @property  # --l3-display-devices
    def l3_display_devices(self) -> bool:
        return bool(self.__settings[TomlSettings.L3_DISPLAY_DEVICES])

    @property  # --include-l3-hosts
    def l3_include_hosts(self) -> bool:
        return bool(self.__settings[TomlSettings.L3_INCLUDE_HOSTS])

    @property  # --l3-skip-ip
    def l3_include_loopback(self) -> bool:
        return bool(self.__settings[TomlSettings.L3_INCLUDE_LOOPBACK])

    @property  # --l3-skip-cidr-0
    def l3_skip_cidr_0(self) -> bool:
        return bool(self.__settings[TomlSettings.L3_SKIP_CIDR_0])

    @property  # --l3-skip-cidr-32-128
    def l3_skip_cidr_32_128(self) -> bool:
        return bool(self.__settings[TomlSettings.L3_SKIP_CIDR_32_128])

    @property  # --l3-skip-if
    def l3_skip_if(self) -> bool:
        return bool(self.__settings[TomlSettings.L3_SKIP_IF])

    @property  # --l3-skip-ip
    def l3_skip_ip(self) -> bool:
        return bool(self.__settings[TomlSettings.L3_SKIP_IP])

    @property  # --l3-skip-public
    def l3_skip_public(self) -> bool:
        return bool(self.__settings[TomlSettings.L3_SKIP_PUBLIC])

    @property  # --layers
    def layers(self) -> List[str]:
        if self.__layers is None:
            self.__layers = list(set(self.__settings[TomlSettings.LAYERS]))
            self.__layers = [entry for entry in self.__layers if entry in Layers.list()]
            if len(self.__layers) != len(self.__settings[TomlSettings.LAYERS]):
                LOGGER.error(f'Wrong/duplicate layer(s) ignored, {self.__settings[TomlSettings.LAYERS]}.')
        return self.__layers

    @property  # --log-file
    def log_file(self) -> str:
        raw_log_file = str(Path(str(self.__settings[TomlSettings.LOG_FILE])).expanduser())
        if not raw_log_file.startswith(f'{LOG_PATH_DEFAULT}/'):
            # logger not ready yet
            print(f'\nInvalid log file {raw_log_file}. Falling back to {LOG_FILE_DEFAULT}')
            return LOG_FILE_DEFAULT
        return raw_log_file

    @property  # --log-level
    def loglevel(self) -> int:
        log_levels = {
            LogLevels.DEBUG: DEBUG,
            LogLevels.INFO: INFO,
            LogLevels.WARNING: WARNING,
            LogLevels.ERROR: ERROR,
            LogLevels.FATAL: FATAL,
            LogLevels.CRITICAL: CRITICAL,
            LogLevels.OFF: -1,
        }
        return log_levels.get(self.__settings[TomlSettings.LOG_LEVEL], WARNING)

    @property  # --log-to-stdout
    def log_to_stdtout(self) -> bool:
        return bool(self.__settings[TomlSettings.LOG_TO_STDOUT])

    @property  # --keep-max-topologies
    def keep_max_topologies(self) -> int | None:
        if isinstance(self.__settings[TomlSettings.KEEP_MAX_TOPOLOGIES], int):
            return self.__settings[TomlSettings.KEEP_MAX_TOPOLOGIES]
        return None

    @property  # --min-topology-age
    def min_topology_age(self) -> int:
        if isinstance(self.__settings[TomlSettings.MIN_TOPOLOGY_AGE], int):
            return max(self.__settings[TomlSettings.MIN_TOPOLOGY_AGE], 0)
        else:
            return 0

    @property  # --output-directory
    def output_directory(self) -> str:
        # init output directory with current time if not set
        if not self.__settings[TomlSettings.OUTPUT_DIRECTORY]:
            self.__settings[TomlSettings.OUTPUT_DIRECTORY] = f'{strftime(self.__settings[TomlSettings.TIME_FORMAT])}'
        raw_output_directory = str(self.__settings[TomlSettings.OUTPUT_DIRECTORY])
        if is_valid_output_directory(raw_output_directory):
            return raw_output_directory
        else:
            LOGGER.error(f'Invalid output directory {raw_output_directory}. Falling back to "nvdct".')
            return 'nvdct'

    @property  # --pre-fetch
    def pre_fetch(self) -> bool:
        return bool(self.__settings[TomlSettings.PRE_FETCH])

    @property  # --quiet
    def quiet(self) -> bool:
        return bool(self.__settings[TomlSettings.QUIET])

    @property  # --time-format
    def time_format(self) -> str:
        return str(self.__settings[TomlSettings.TIME_FORMAT].replace('/', '_').replace('..', '__'))

    @property  # --update-config
    def update_config(self) -> bool:
        return bool(self.__settings[TomlSettings.UPDATE_CONFIG])

    #
    #  user data setting
    #

    @property
    def emblems(self) -> Emblems:
        if self.__emblems is None:
            raw_emblems = self.__user_data.get(TomlSections.EMBLEMS, {})
            self.__emblems = Emblems(
                host_node=str(raw_emblems.get(EmblemNames.HOST_NODE, EmblemValues.ICON_ALERT_UNREACHABLE)),
                ip_address=str(raw_emblems.get(EmblemNames.IP_ADDRESS, EmblemValues.IP_ADDRESS_80)),
                ip_network=str(raw_emblems.get(EmblemNames.IP_NETWORK, EmblemValues.IP_NETWORK_80)),
                l3_replace=str(raw_emblems.get(EmblemNames.L3_REPLACE, EmblemValues.ICON_PLUGINS_CLOUD)),
                l3_summarize=str(raw_emblems.get(EmblemNames.L3_SUMMARIZE, EmblemValues.ICON_AGGREGATION)),
                service_node=str(raw_emblems.get(EmblemNames.SERVICE_NODE, EmblemValues.ICON_ALERT_UNREACHABLE)),
            )
        return self.__emblems

    @property
    def filter_by_customer(self) -> List[str]:
        if self.__filter_by_customer is None:
            self.__filter_by_customer = [
                str(customer) for customer in set(self.__user_data.get(TomlSections.FILTER_BY_CUSTOMER, []))
                if is_valid_customer_name(customer)]
            LOGGER.info(f'Found {len(self.__filter_by_customer)} customer(s) to filter on')
        return self.__filter_by_customer

    @property
    def filter_by_site(self) -> List[str]:
        if self.__filter_by_site is None:
            self.__filter_by_site = [str(site) for site in set(self.__user_data.get(TomlSections.FILTER_BY_SITE, [])) if is_valid_site_name(site)]
            LOGGER.info(f'Found {len(self.__filter_by_site)} site(s) to filter on')
        return self.__filter_by_site

    @staticmethod
    def parse_key_value_section(section: str, data: Mapping[str, str]) -> Mapping[str, set[str]]:
        parsed = {
            IncludeExclude.INCLUDE: set(),
            IncludeExclude.EXCLUDE: set()
        }
        for key_value, mode in data.items():
            if mode not in IncludeExclude.list():
                LOGGER.error(
                    f'Invalid mode in {section} found: {key_value}={mode} -> line ignored'
                )
                continue
            match section:
                case TomlSections.FILTER_BY_HOST_LABEL:
                    try:
                        key, value = key_value.split(':', 1)
                    except ValueError:
                        LOGGER.error(
                            f'Invalid host label found missing ":": {key_value}={mode} -> line ignored'
                        )
                        continue
                    if ':' in value:
                        LOGGER.error(
                            f'Invalid host label found can not contain more than one ":": "{key_value}={mode}" -> line ignored'
                        )
                        continue
                    parsed[mode].add(f"'{key}' '{value}'")
                case TomlSections.FILTER_BY_FOLDER:
                    parsed[mode].add(f'^/wato/{key_value.strip("/")}/')
                case _:
                    parsed[mode].add(key_value)
        return parsed

    @property
    def filter_by_folder(self) -> Mapping[str, set[str]]:
        if self.__filter_by_folder is None:
            self.__filter_by_folder = self.parse_key_value_section(
                section=TomlSections.FILTER_BY_FOLDER,
                data=self.__user_data.get(TomlSections.FILTER_BY_FOLDER, {})
            )
        return self.__filter_by_folder

    @property
    def filter_by_host_label(self) -> Dict[str, Set[str]]:
        if self.__filter_by_host_label is None:
            self.__filter_by_host_label = self.parse_key_value_section(
                section=TomlSections.FILTER_BY_HOST_LABEL,
                data=self.__user_data.get(TomlSections.FILTER_BY_HOST_LABEL, {})
            )
        return self.__filter_by_host_label

    @property
    def filter_by_host_tag(self) -> Dict[str, Set[str]]:
        if self.__filter_by_host_tag is None:
            self.__filter_by_host_tag = self.parse_key_value_section(
                section=TomlSections.FILTER_BY_HOST_TAG,
                data=self.__user_data.get(TomlSections.FILTER_BY_HOST_TAG, {})
            )
        return self.__filter_by_host_tag

    @property
    def l2_drop_neighbours(self) -> List[str]:
        if self.__l2_drop_neighbours is None:
            self.__l2_drop_neighbours = [str(host) for host in set(self.__user_data.get(TomlSections.L2_DROP_NEIGHBOURS, []))]
        return self.__l2_drop_neighbours

    @property
    def l2_seed_devices(self) -> List[str]:
        if self.__l2_seed_devices is None:
            self.__l2_seed_devices = list(set(str(host) for host in (
                self.__user_data.get(TomlSections.L2_SEED_DEVICES, [])) if is_valid_hostname(host)))
        return self.__l2_seed_devices

    @property
    def l2_neighbour_to_host_map(self) -> Dict[str, str]:
        if self.__l2_neighbour_to_host_map is None:
            self.__l2_neighbour_to_host_map = {
                str(neighbour): str(host) for neighbour, host in self.__user_data.get(
                    TomlSections.L2_NEIGHBOUR_TO_HOST_MAP, {}
                ).items() if is_valid_hostname(neighbour)
            }
        return self.__l2_neighbour_to_host_map

    @property
    def l2_neighbour_replace_regex(self) -> List[Tuple[str, str]] | None:
        if self.__l2_neighbour_replace_regex is None:
            self.__l2_neighbour_replace_regex = [
                (
                    str(regex), str(replace)
                ) for regex, replace in self.__user_data.get(TomlSections.L2_NEIGHBOUR_REPLACE_REGEX, {}).items()
            ]
        return self.__l2_neighbour_replace_regex

    @property
    def l3_ignore_hosts(self) -> List[str]:
        if self.__l3_ignore_hosts is None:
            self.__l3_ignore_hosts = [str(host) for host in set(self.__user_data.get(
                TomlSections.L3_IGNORE_HOSTS, []
            )) if is_valid_hostname(host)]
        return self.__l3_ignore_hosts

    @property
    def l3_ignore_ips(self) -> List[ip_network]:
        if self.__l3_ignore_ip is None:
            self.__l3_ignore_ip = []
            for raw_ip_network in self.__user_data.get(TomlSections.L3_IGNORE_IP, []):
                try:
                    self.__l3_ignore_ip.append(ip_network(raw_ip_network, strict=False))
                except (AddressValueError, NetmaskValueError, ValueError):
                    LOGGER.error(
                        f'Invalid entry in {TomlSections.L3_IGNORE_IP} found: {raw_ip_network} -> ignored'
                    )
                    continue
            LOGGER.info(
                f'Valid entries in {TomlSections.L3_IGNORE_IP} found: {len(self.__l3_ignore_ip)}/'
                f'{len(self.__user_data.get(TomlSections.L3_IGNORE_IP, []))}'
            )
        return self.__l3_ignore_ip

    @property
    def l3v4_ignore_wildcard(self) -> List[Wildcard]:
        if self.__l3v4_ignore_wildcard is None:
            self.__l3v4_ignore_wildcard = []
            for entry in self.__user_data.get(TomlSections.L3V4_IGNORE_WILDCARD, []):
                try:
                    raw_ip_address, wildcard = entry
                except ValueError:
                    LOGGER.error(
                        f'Invalid entry in {TomlSections.L3V4_IGNORE_WILDCARD} -> {entry} -> ignored'
                    )
                    continue
                try:
                    # 0.0.127.0 -> 255.255.128.255
                    inverted_wildcard = '.'.join(
                        [str(255 - int(octet)) for octet in wildcard.split('.')]
                    )

                    self.__l3v4_ignore_wildcard.append(Wildcard(
                        int_ip_address=int(ip_address(raw_ip_address)),
                        int_wildcard=int(ip_address(inverted_wildcard)),
                        ip_address=raw_ip_address,
                        wildcard=wildcard,
                        bit_pattern=int(ip_address(raw_ip_address)) & int(
                            ip_address(inverted_wildcard)
                        )
                    ))
                except (AddressValueError, NetmaskValueError, ValueError):
                    LOGGER.error(
                        f'Invalid entry in {TomlSections.L3V4_IGNORE_WILDCARD} -> {entry} -> ignored'
                    )
                    continue
            LOGGER.info(
                f'Valid entries in {TomlSections.L3V4_IGNORE_WILDCARD} found: {len(self.__l3v4_ignore_wildcard)}/'
                f'{len(self.__user_data.get(TomlSections.L3V4_IGNORE_WILDCARD, []))}'
            )
        return self.__l3v4_ignore_wildcard

    @property
    def l3_replace_networks(self) -> Dict[str, str]:
        if self.__l3_replace is None:
            self.__l3_replace = {}
            for raw_ip_network, node in self.__user_data.get(TomlSections.L3_REPLACE_NETWORKS, {}).items():
                try:
                    _ip_network = ip_network(raw_ip_network)  # noqa: F841
                except (AddressValueError, NetmaskValueError, ValueError):
                    LOGGER.error(
                        f'Invalid entry in {TomlSections.L3_REPLACE_NETWORKS} found: "{raw_ip_network}" = "{node}" -> line ignored'
                    )
                    continue
                if not is_valid_hostname(node):
                    LOGGER.error(f'Invalid node name found: {node} -> line ignored')
                    continue
                self.__l3_replace[raw_ip_network] = str(node)
            LOGGER.info(
                f'Valid entries in {TomlSections.L3_REPLACE_NETWORKS} found: {len(self.__l3_replace)}/'
                f'{len(self.__user_data.get(TomlSections.L3_REPLACE_NETWORKS, {}))}'
            )
        return self.__l3_replace

    @property
    def l3_summarize(self) -> List[ip_network]:
        if self.__l3_summarize is None:
            self.__l3_summarize = []
            for raw_ip_network in self.__user_data.get(TomlSections.L3_SUMMARIZE, []):
                try:
                    self.__l3_summarize.append(ip_network(raw_ip_network, strict=False))
                except (AddressValueError, NetmaskValueError, ValueError):
                    LOGGER.error(
                        f'Invalid entry in {TomlSections.L3_SUMMARIZE} -> {raw_ip_network} -> ignored'
                    )
                    continue
            LOGGER.info(
                f'Valid entries in {TomlSections.L3_SUMMARIZE} found: {len(self.__l3_summarize)}/'
                f'{len(self.__user_data.get(TomlSections.L3_SUMMARIZE, []))}'
            )
        return self.__l3_summarize

    @property
    def map_speed_to_thickness(self) -> List[Thickness]:
        if self.__map_speed_to_thickness is None:
            self.__map_speed_to_thickness = []
            map_speed_to_thickness = self.__user_data.get(
                TomlSections.MAP_SPEED_TO_THICKNESS, {}
            )
            for speed, thickness in map_speed_to_thickness.items():
                try:
                    self.__map_speed_to_thickness.append(Thickness(
                        speed=int(float(speed)),
                        thickness=int(thickness),
                    ))
                except ValueError:
                    LOGGER.error(
                        f'Invalid entry in {TomlSections.MAP_SPEED_TO_THICKNESS} -> {speed}={thickness} -> ignored'
                    )
                    continue
            LOGGER.info(
                f'Valid entries in {TomlSections.MAP_SPEED_TO_THICKNESS} found: {len(self.__map_speed_to_thickness)}'
                f'/{len(self.__user_data.get(TomlSections.MAP_SPEED_TO_THICKNESS, []))}'
            )
        return self.__map_speed_to_thickness

    @property
    def protected_topologies(self) -> List[str]:
        if self.__protected_topologies is None:
            self.__protected_topologies = [str(topology) for topology in self.__user_data.get(
                TomlSections.PROTECTED_TOPOLOGIES, []
            )]
        return self.__protected_topologies

    @property
    def static_connections(self) -> List[StaticConnection]:
        if self.__static_connections is None:
            self.__static_connections = []
            for connection in self.__user_data.get(TomlSections.STATIC_CONNECTIONS, []):
                try:
                    left_host, left_service, right_service, right_host = connection
                except ValueError:
                    LOGGER.error(
                        f'Invalid entry in {TomlSections.STATIC_CONNECTIONS} needs to be 4 tuples: -> {connection} -> ignored'
                    )
                    continue
                if not right_host or not left_host:
                    LOGGER.warning(f'Both hosts must be set, got: {connection}  -> ignored')
                    continue
                if not is_valid_hostname(right_host) or not is_valid_hostname(left_host):
                    continue
                self.__static_connections.append(StaticConnection(
                    right_host=str(right_host),
                    right_service=str(right_service),
                    left_service=str(left_service),
                    left_host=str(left_host),
                ))
            LOGGER.info(
                f'Valid entries in {TomlSections.STATIC_CONNECTIONS} found: {len(self.__static_connections)}/'
                f'{len(self.__user_data.get(TomlSections.STATIC_CONNECTIONS, []))}'
            )
        return self.__static_connections


