#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2

# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2024-06-09
# File  : nvdct/lib/topologies.py

# 2024-12-22: refactoring topology creation into classes
#             made L3 topology IP version independent
# 2024-12-25: refactoring, moved function into classes
# 2025-01-22: changed: show interface service in L2 and L3 topology instead of Port/Device from HW/SW inventory

from abc import abstractmethod
from collections.abc import Mapping, MutableMapping, Sequence, MutableSet
from ipaddress import ip_address, ip_interface, ip_network
from re import sub as re_sub
from typing import Dict, List, Tuple

from lib.backends import (
    CacheItems,
    HostCache,
)
from lib.constants import (
    CACHE_INTERFACES_DATA,
    HostFilter,
    HostLabels,
    IPVersion,
    InvPaths,
    L2IgnoreMismatch,
    L2InvColumns,
    L3InvColumns,
    LOGGER,
    LiveStatusOperator,
    TomlSections,
)
from lib.settings import (
    Emblems,
    StaticConnection,
    Thickness,
    Wildcard,
)
from lib.utils import (
    save_data_to_file,
)


class NvObjects:
    def __init__(
            self,
            host_cache: HostCache
    ) -> None:
        self.nv_objects: Dict[str, any] = {}
        self.host_count: int = 0
        self.host_list: List[str] = []
        self.host_cache = host_cache

    def add_host(
            self,
            host: str,
            emblem: str | None = None,
            display_name: str | None = None,
    ) -> None:
        """
        Adds a host object to the topology.
        - if there is no display name, "host" will be used as display_name
        - the emblem for the host will only be used if the host is not a Checkmk host object
        - if there is no "emblem" the Checkmk will use its default emblem for missing objects (white question mark on blue ground)
        Args:
            host: the name of the host
            emblem: the emblem for the object
            display_name: the name for the host to show in the topology
        Returns:
            None
        """
        if display_name and host in self.nv_objects:
            self.nv_objects[host]['name'] = display_name

        if host not in self.nv_objects:
            self.host_count += 1
            self.host_list.append(host)
            link: Dict = {}
            metadata: Dict = {}
            # LOGGER.debug(f'host: {host}, {host_cache.host_exists(host=host)}')
            if self.host_cache.host_exists(host=host) is True:
                LOGGER.debug(f'host exists : {host}')
                link = {'core': host}
            else:
                if emblem is not None:
                    metadata = {
                        'images': {
                            'emblem': emblem,  # node image
                            # 'icon': 'icon_tick',  # status image, top left from emblem
                        },
                        **({'tooltip': {
                            'quickinfo': [{'name': 'Host node', 'value': host}]
                        }} if not link else {})
                    }

            self.nv_objects[host] = {
                'name': display_name if display_name is not None else host,
                'link': link,
                'metadata': metadata,
            }
            LOGGER.debug(f'host: {host}, link: {link}, metadata: {metadata}')

    def add_service(
            self,
            host: str,
            service: str,
            emblem: str | None = None,
            metadata: Dict | None = None,
            display_name: str | None = None,
    ) -> None:
        """
        Adds a service object to the topology.
        - the service object will be added as "service@host" to the topology
        - service should be the complete service name as in Checkmk, i.e. "Interface X440G2-48p-10G4 Port 52"
        - if the host is a Checkmk host object it will be linked to the core
        - if there is no display name, "service" will be used as display_name
        - the emblem for the service will only be used if the host is not a Checkmk host object
        - if there is no "emblem" the Checkmk will use its default emblem for missing objects (white question mark on blue ground)
        Args:
            host: the name of the host
            service: the name of the service
            emblem: the emblem for the object
            metadata: additional data for the service
            display_name: the name for the host to show in the topology
        Returns:
            None
        """
        if metadata is None:
            metadata = {}
        if display_name is None:
            display_name = service

        self.add_host(host=host)
        service_object = f'{service}@{host}'
        if service_object not in self.nv_objects:
            link: Dict = {}
            if self.host_cache.host_exists(host=host):
                link = {'core': [host, service]}
            elif emblem is not None:
                metadata.update({
                    'images': {
                        'emblem': emblem,  # node image
                        # 'icon': 'icon_tick',  # status image, top left from emblem
                    },
                    **({'tooltip': {'quickinfo': [{'name': 'Service node', 'value': service}]}} if not link else {})
                })

            self.nv_objects[service_object] = {
                'name': display_name,
                'link': link,
                'metadata': metadata,
            }
        elif metadata is not {}:
            self.nv_objects[service_object]['metadata'].update(metadata)

    def add_interface(
            self,
            host: str,
            service: str,
            item: str | None,
            emblem: str | None = None,
            metadata: Dict | None = None,
            display_name: str | None = None,
    ) -> None:
        """
        Adds an interface service object to the topology.
        - the interface service object will be added as "service@host" to the topology
        - service ist the interface name from the H/W-inventory
        - "item" should be the interface item as in Checkmk, i.e. "X440G2-48p-10G4 Port 52"
        - if there is no "item" service will be used as "item"
        - if the host is a Checkmk host object it will be linked to the core
        - if there is no display name, "service" will be used as display_name
        - the "emblem" for the service will only be used if the host is not a Checkmk host object
        - if there is no "emblem" the Checkmk will use its default emblem for missing objects (white question mark on blue ground)
        Args:
            host: the name of the host
            service: the name of the service
            item: the interface name as in Checkmk
            emblem: the emblem for the object
            metadata: additional data for the service
            display_name: the name for the host to show in the topology
        Returns:
            None
        """
        if metadata is None:
            metadata = {}
        if display_name is None:
            display_name = service
        speed = None

        self.add_host(host=host)
        service_object = f'{service}@{host}'
        if service_object not in self.nv_objects:
            link: Dict = {}
            # if item is None:
            #     item = get_service_by_interface(host, service.lstrip('0'), host_cache)
            if item and self.host_cache.host_exists(host=host):
                service_long = f'Interface {item}'
                link = {'core': [host, service_long]}
                if op_data := self.get_operational_interface_data(host, item):
                    metadata.update(op_data)
                    speed = op_data.get('op_speed_int')
            elif emblem is not None:
                metadata.update({
                    'images': {
                        'emblem': emblem,  # node image
                        # 'icon': 'icon_tick',  # status image, top left from emblem
                    },
                    **({'tooltip': {'quickinfo': [
                        {'name': 'Service node', 'value': item}
                    ]}} if not link else {})
                })
            if isinstance(speed, int):
                metadata.update({'native_speed': speed})

            self.nv_objects[service_object] = {
                'name': display_name,
                'link': link,
                'metadata': metadata,
            }
        elif metadata is not {}:
            self.nv_objects[service_object]['metadata'].update(metadata)

    def add_ip_address(
            self,
            host: str,
            raw_ip_address: str,
            emblem: str,
            interface: str | None,
    ) -> None:
        """
        Adds an ip address object to the topology.
        - the ip address object will be added as "raw_ip_address@interface@host" to the topology
        - if interface is None (--l2-skip-if) the object is added as "raw_ip_address@host"
        Args:
            host: the name of the Checkmk host
            raw_ip_address: the ip address as string (i.e. "10.10.10.10")
            emblem: the emblem for the object
            interface: the interface of the host where the ip address belongs to
        Returns:
            None
        """

        if interface is not None:
            service_object = f'{raw_ip_address}@{interface}@{host}'
        else:
            service_object = f'{raw_ip_address}@{host}'

        if service_object not in self.nv_objects:
            self.nv_objects[service_object] = {
                'name': raw_ip_address,
                'link': {},
                'metadata': {
                    'images': {
                        'emblem': emblem,
                        # 'icon': 'icon_tick',
                    },
                    'tooltip': {
                        'quickinfo': [
                            {'name': 'IP-Address', 'value': raw_ip_address},
                        ]
                    }
                }
            }

    def get_operational_interface_data(
            self,
            host: str,
            item: str,
    ) -> Dict[str, str | int] | None:
        unit_to_bits_per_second = {
            'Bit/s': 1,
            'kBit/s': 1000,
            'Kbps': 1000,
            'MBit/s': 1000000,
            'Mbps': 1000000,
            'GBit/s': 1000000000,
            'Gbps': 1000000000,
        }

        # get dict of interfaces with the item as key
        interface_data: Dict[str, any] | None = self.host_cache.get_data(
            host=host, item=CacheItems.interfaces, path=CACHE_INTERFACES_DATA
        )
        try:
            raw_operational_data = interface_data[item]['long_plugin_output']
        except (KeyError, TypeError):
            return None

        if raw_operational_data:
            operational_data: Dict[str, str | int] = {}
            for _entry in raw_operational_data:
                try:
                    key, value = _entry.split(': ', 1)  # split only at the first colon
                except ValueError:
                    continue
                value = value.strip(' ')
                match key:
                    case 'MAC':
                        if len(value) == 17:  # valid MAC: 6C:DD:30:DD:51:8B'
                            operational_data['mac'] = value
                    case 'Speed':
                        try:  # *_ -> ignore rest of string, i.e: (expected: 1 GBit/s)WARN
                            speed, unit, *_ = value.split(' ')
                        except ValueError:
                            pass
                        else:
                            operational_data['op_sped_str'] = f'{speed} {unit}'
                            operational_data['op_speed_int'] = int(float(speed) * unit_to_bits_per_second[unit])

            return operational_data
        return None

    def add_ip_network(
            self,
            network: str,
            emblem: str,
    ) -> None:
        """
        Adds an ip network object to the topology.
        - the ip network object will be added as "network" to the topology
        Args:
            network: the network as string (i.e. "10.10.10.0/24")
            emblem: the emblem for the object
        Returns:
            None
        """
        if network not in self.nv_objects:
            self.nv_objects[network] = {
                'name': network,
                'link': {},
                'metadata': {
                    'images': {
                        'emblem': emblem,  # node image
                        # 'icon': 'icon_tick',  # status image, top left from emblem
                    },
                    'tooltip': {
                        'quickinfo': [
                            {'name': 'IP-Network', 'value': network},
                        ]
                    }
                }
            }

    def add_tooltip_quickinfo(self, nv_object: str, display_name: str, value: str) -> Dict:
        metadata = self.nv_objects[nv_object]['metadata']
        if metadata.get('tooltip') is None:
            metadata['tooltip'] = {}
        if metadata['tooltip'].get('quickinfo') is None:
            metadata['tooltip']['quickinfo'] = []

        metadata['tooltip']['quickinfo'].append({
            'name': display_name,
            'value': value,
        })

        return metadata

    def add_icon_to_object(self, nv_object: str, icon: str) -> None:
        if nv_object in self.nv_objects:
            if not self.nv_objects[nv_object]['metadata'].get('images'):
                self.nv_objects[nv_object]['metadata']['images'] = {}
            self.nv_objects[nv_object]['metadata']['images']['icon'] = icon


class NvConnections:
    def __init__(self) -> None:
        self.nv_connections: List = []

    def add_connection(self, left: str, right: str):
        connection = [left, right]
        connection.sort()
        if [connection] not in self.nv_connections:
            self.nv_connections.append([connection])

    def add_meta_data_to_connections(
            self,
            nv_objects: NvObjects,
            speed_map: Sequence[Thickness],
            ignore_mismatch: Sequence[str],
    ):
        for connection in self.nv_connections:
            warning = False
            left, right = connection[0]
            metadata: Dict[str, Dict] = {
                'line_config': {
                    'css_styles': {
                        'stroke-dasharray': 'unset'
                    }
                }
            }

            left_speed = nv_objects.nv_objects[left].get('metadata', {}).get('op_speed_int')
            right_speed = nv_objects.nv_objects[right].get('metadata', {}).get('op_speed_int')

            left_speed_str = nv_objects.nv_objects[left].get('metadata', {}).get('op_sped_str')
            right_speed_str = nv_objects.nv_objects[right].get('metadata', {}).get('op_sped_str')

            left_duplex = nv_objects.nv_objects[left].get('metadata', {}).get('duplex')
            right_duplex = nv_objects.nv_objects[right].get('metadata', {}).get('duplex')

            left_native_vlan = nv_objects.nv_objects[left].get('metadata', {}).get('native_vlan')
            right_native_vlan = nv_objects.nv_objects[right].get('metadata', {}).get('native_vlan')

            if right_speed and left_speed:
                right_thickness = map_speed_to_thickness(right_speed, speed_map)
                # left_thickness = map_speed_to_thickness(left_speed, speed_map)
                metadata['line_config']['thickness'] = right_thickness

                if right_speed != left_speed and not L2IgnoreMismatch.SPEED in ignore_mismatch:
                    warning = True
                    metadata = add_tooltip_html(
                        metadata, 'Speed', left, left_speed_str, right, right_speed_str
                    )
                    # metadata = add_tooltip_quickinfo(metadata, 'Warning', 'Speed mismatch')
                    # metadata = add_tooltip_quickinfo(metadata, left, left_speed_str)
                    # metadata = add_tooltip_quickinfo(metadata, right, right_speed_str)

                    LOGGER.warning(
                        f'Connection speed mismatch: {left} ({left_speed_str})'
                        f'<->{right} ({right_speed_str})'
                    )

            # for duplex/native vlan it might be a good idea to change left/right
            # value as they are reported by CDP as neighbour states
            if left_duplex and right_duplex:
                if left_duplex != right_duplex and not L2IgnoreMismatch.DUPLEX in ignore_mismatch:
                    warning = True

                    metadata = add_tooltip_html(
                        metadata, 'Duplex', left, left_duplex, right, right_duplex
                    )

                    LOGGER.warning(
                        f'Connection duplex mismatch: {left} ({left_duplex})'
                        f'<->{right} ({right_duplex})'
                    )
            if left_native_vlan and right_native_vlan and not L2IgnoreMismatch.VLAN in ignore_mismatch:
                if left_native_vlan != '0' and right_native_vlan != '0':  # ignore VLAN 0 (Native VLAN on routed ports)
                    if left_native_vlan != right_native_vlan:
                        warning = True

                        metadata = add_tooltip_html(
                            metadata, 'Native<br>VLAN', left, left_native_vlan, right, right_native_vlan
                        )

                        LOGGER.warning(
                            f'Connection native vlan mismatch: '
                            f'{left} ({left_native_vlan})<->{right} ({right_native_vlan})'
                        )
            if warning:
                metadata['line_config'].update({
                    'color': 'red',
                    'thickness': 5,
                })
                metadata['line_config']['css_styles']['stroke-dasharray'] = '10 5'
                nv_objects.add_icon_to_object(left, 'icon_warning')
                nv_objects.add_icon_to_object(right, 'icon_warning')
                close_tooltip_html(metadata)
            connection.append(metadata)


class Topology:
    def __init__(
            self,
            emblems: Emblems,
            host_cache: HostCache,
            topology: str,
    ):
        self.nv_objects: NvObjects = NvObjects(host_cache=host_cache)
        self.nv_connections: NvConnections = NvConnections()
        self.emblems: Emblems = emblems
        self.host_cache: HostCache = host_cache
        self.topology = topology

    @abstractmethod
    def create(self):
        raise NotImplementedError

    def save(self, label: str, output_directory: str, make_default: bool, dont_compare):
        LOGGER.info(f'{self.topology} saving...')
        data = {
            'version': 1,
            'name': label,
            'objects': dict(sorted(self.nv_objects.nv_objects.items())),
            'connections': sorted(self.nv_connections.nv_connections)
        }
        save_data_to_file(
            data=data,
            path=output_directory,
            file=f'data_{label}.json',
            make_default=make_default,
            dont_compare=dont_compare,
        )

    def get_service_by_interface(self, host: str, interface: str) -> str | None:
        """
        Returns:
            Tuple of interface item
        """

        short_if_names = {
            'ethernet': 'eth',
            # 'fastethernet': 'Fa',
            # 'gigabitethernet': 'gi',
            # 'tengigabitethernet': 'te',
            # 'fortygigabitethernet': 'Fo',
            # 'hundredgigabitethernet': 'Hu',
            # 'management': 'Ma',
        }

        def get_short_if_name(long_interface: str) -> str:
            """
            returns short interface name from long interface name
            interface: is the long interface name
            :type long_interface: str
            """
            if not long_interface:
                return long_interface
            for interface_prefix in short_if_names.keys():
                if long_interface.lower().startswith(interface_prefix.lower()):
                    interface_short = short_if_names[interface_prefix]
                    return long_interface.lower().replace(interface_prefix.lower(), interface_short, 1)
            return long_interface

        # try to find the item for an interface
        def match_entry_with_item(interface_entry: Mapping[str, str], services: Sequence[str]) -> str | None:
            values = [
                interface_entry.get('name', '').strip(),
                interface_entry.get('description','').strip(),
                interface_entry.get('alias', '').strip()
            ]
            for value in values:
                if value in services:
                    return value

            index = str(interface_entry.get('index'))

            # try alias+index
            alias_index = str(interface_entry.get('alias')).strip() + ' ' + index
            if alias_index in services:
                LOGGER.info(f'{self.topology} match found by alias-index |{interface_entry}| <-> |{alias_index}|')
                return alias_index

            # try description+index
            description_index = str(interface_entry.get('description')).strip() + ' ' + index
            if description_index in services:
                LOGGER.info(f'{self.topology} match found by description-index |{interface_entry}| <-> |{description_index}|')
                return description_index

            # for index try with padding
            pad_services: List[str] = [x for x in services if x.isdigit()]
            if pad_services:
                max_pad: int = len(max(pad_services, key=len)) + 1
                # min_pad: int = len(min(pad_services, key=len))
                min_pad: int = len(min(pad_services, key=len))
                for i in range(min_pad, max_pad):
                    index_padded = f'{index:0>{i}}'
                    if index_padded in pad_services:
                        return index_padded
                    # still not found try values + index
                    for value in values:
                        if f'{value} {index_padded}' in services:
                            return f'{value} {index_padded}'

            LOGGER.warning(f'{self.topology} no match found |{interface_entry} | <-> |{services}|')
            return None

        # empty host/neighbour should never happen here
        if not host:
            LOGGER.warning(f'{self.topology} no host name |{host}|')
            return None

        # get dict of interfaces with the item as key
        interface_data: Mapping[str, Mapping[str, object]] = self.host_cache.get_data(
            host=host, item=CacheItems.interfaces, path=CACHE_INTERFACES_DATA
        )
        if not interface_data:
            LOGGER.warning(f'{self.topology} no interface data for: {host}')
            return None

        # try to find the interface in the host interface inventory list
        inventory = self.host_cache.get_data(
            host=host, item=CacheItems.inventory, path=InvPaths.INTERFACES
        )
        if not inventory:
            LOGGER.warning(f'{self.topology} no interface inventory for: {host}')
            return None

        interface_items: Sequence[str] = list(interface_data.keys())

        # the easy case
        if interface in interface_items:
            return interface

        for entry in inventory:
            if interface in [
                entry.get('name'),
                entry.get('description'),
                entry.get('alias'),
                str(entry.get('index')),
                entry.get('phys_address'),
            ]:
                return match_entry_with_item(entry, interface_items)
            elif f'1:{interface}' == entry.get('name'):  # Extreme non stack:
                return match_entry_with_item(entry, interface_items)
            elif entry.get('name') is not None and get_short_if_name(
                    entry.get('name')) == str(interface).lower():  # Cisco NXOS
                return match_entry_with_item(entry, interface_items)

        LOGGER.warning(f'{self.topology} Service for interface not found: {host}, |{interface}|')


class TopologyStatic(Topology):
    def __init__(
            self,
            connections: Sequence[StaticConnection],
            emblems: Emblems,
            host_cache: HostCache,
    ):
        super().__init__(
            emblems=emblems,
            host_cache=host_cache,
            topology='[STATIC]',
        )
        self.connections: Sequence[StaticConnection] = connections

    def create(self):
        for connection in self.connections:
            LOGGER.debug(f'{self.topology} connection from {TomlSections.STATIC_CONNECTIONS}: {connection}')
            self.nv_objects.add_host(
                host=connection.right_host,
                emblem=self.emblems.host_node
            )
            self.nv_objects.add_host(
                host=connection.left_host,
                emblem=self.emblems.host_node
            )
            if connection.right_service:
                self.nv_objects.add_service(
                    host=connection.right_host,
                    emblem=self.emblems.service_node,
                    service=connection.right_service
                )
                self.nv_connections.add_connection(
                    left=connection.right_host,
                    right=f'{connection.right_service}@{connection.right_host}',
                )

            if connection.left_service:
                self.nv_objects.add_service(
                    host=connection.left_host,
                    emblem=self.emblems.service_node,
                    service=connection.left_service
                )
                self.nv_connections.add_connection(
                    left=connection.left_host,
                    right=f'{connection.left_service}@{connection.left_host}',
                )

            if connection.right_service and connection.left_service:
                self.nv_connections.add_connection(
                    left=f'{connection.right_service}@{connection.right_host}',
                    right=f'{connection.left_service}@{connection.left_host}',
                )
            elif connection.right_service:  # connect right_service with left_host
                self.nv_connections.add_connection(
                    left=f'{connection.right_service}@{connection.right_host}',
                    right=f'{connection.left_host}',
                )
            elif connection.left_service:  # connect left_service with right_host
                self.nv_connections.add_connection(
                    left=f'{connection.right_host}',
                    right=f'{connection.left_service}@{connection.left_host}',
                )
            else:  # connect right_host with left_host
                self.nv_connections.add_connection(
                    left=f'{connection.right_host}',
                    right=f'{connection.left_host}',
                )


class TopologyL2(Topology):
    def __init__(
            self,
            display_neighbours: bool,
            display_ports: bool,
            drop_neighbours: List[str],
            emblems: Emblems,
            host_cache: HostCache,
            label: str,
            neighbour_replace_regex: List[Tuple[str, str]],
            path_in_inventory: str,
            seed_devices: Sequence[str],
            skip_external: bool,
    ):
        super().__init__(
            emblems=emblems,
            host_cache=host_cache,
            topology = f'[L2 {label}]',
        )
        self.display_neighbours: bool = display_neighbours
        self.display_ports: bool = display_ports
        self.drop_neighbours: List[str] = drop_neighbours
        self.hosts_done: MutableSet[str] = set()
        self.hosts_to_go: MutableSet[str] = set(seed_devices)
        self.label: str = label
        self.neighbour_replace_regex: List[Tuple[str, str]] = neighbour_replace_regex
        self.neighbour_to_host: MutableMapping[str, str | None] = {}
        self.path_in_inventory: str = path_in_inventory
        self.raw_neighbour_to_neighbour: Dict[str, str] = {}
        self.skip_external: bool = skip_external

    def create(self):
        if not self.hosts_to_go:
            LOGGER.error(f'{self.topology} no seed devices !')
            return

        while self.hosts_to_go:
            host = self.hosts_to_go.pop()
            self.hosts_done.add(host)
            if not self.host_cache.is_host_allowed(host):
                LOGGER.info(f'{self.topology} host dropped by filter: {host}')
                continue

            topo_data: Sequence[Mapping[str, str]] = self.host_cache.get_data(
                host=host, item=CacheItems.inventory, path=self.path_in_inventory
            )
            if topo_data:
                self.host_from_inventory(
                    host=host,
                    inv_data=topo_data,
                )

            LOGGER.info(f'{self.topology} host done: {host}')

    def host_from_inventory(
            self,
            host: str,
            inv_data: Sequence[Mapping[str,str]],
    ):
        for topo_neighbour in inv_data:
            # check if required data are not empty
            if not (raw_neighbour := topo_neighbour.get(L2InvColumns.NEIGHBOUR)):
                LOGGER.warning(f'{self.topology} incomplete data neighbour missing: {topo_neighbour}')
                continue

            # stop here if neighbour is dropped anyway...
            if not (neighbour := self.adjust_raw_neighbour(raw_neighbour)):
                continue

            if not (raw_local_port := topo_neighbour.get(L2InvColumns.LOCALPORT)):
                LOGGER.warning(f'{self.topology} incomplete data local port missing: {topo_neighbour}')
                continue

            if not (raw_neighbour_port := topo_neighbour.get(L2InvColumns.NEIGHBOURPORT)):
                LOGGER.warning(f'{self.topology} incomplete data neighbour port missing: {topo_neighbour}')
                continue

            # get local interface service
            if not (local_port := self.get_service_by_interface(host, raw_local_port)):
                local_port = raw_local_port
                LOGGER.warning(f'{self.topology} service not found for local_port: {host}, {raw_local_port}')
            elif local_port != raw_local_port:
                LOGGER.info(
                    f'{self.topology} map raw_local_port -> local_port: {host}, {raw_local_port} -> {local_port}'
                )

            if neighbour_host := self.host_cache.get_host_from_neighbour(
                    neighbour=neighbour,
                    neighbour_id=topo_neighbour.get(L2InvColumns.NEIGHBOURID),
                    raw_neighbour=raw_neighbour,
            ):
                if not self.host_cache.is_host_allowed(neighbour_host):
                    LOGGER.info(f'{self.topology} neighbour dropped by include/exclude filter: {neighbour_host}')
                    continue

                if neighbour_host not in self.hosts_done:
                    self.hosts_to_go.add(neighbour_host)

                if not (neighbour_port := self.get_service_by_interface(neighbour_host, raw_neighbour_port)):
                    neighbour_port = raw_neighbour_port
                    LOGGER.warning(
                        f'{self.topology} service not found for neighbour port:  {neighbour_host}, {raw_neighbour_port}'
                    )
                elif neighbour_port != raw_neighbour_port:
                    # neighbour_port = raw_neighbour_port  # don't reset neighbour_port
                    LOGGER.info(
                        f'{self.topology} map raw_neighbour_port -> '
                        f'neighbour_port: {neighbour_host}, {raw_neighbour_port} -> {neighbour_port}'
                    )
                neighbour_name = raw_neighbour if self.display_neighbours else None
            else:
                # neighbour is external to cmk, use neighbour id if available as neighbour
                if self.skip_external:
                    continue
                if not (neighbour_host := topo_neighbour.get(L2InvColumns.NEIGHBOURID)):
                    neighbour_host = neighbour
                neighbour_port = raw_neighbour_port
                neighbour_name = raw_neighbour if self.display_neighbours else neighbour
            metadata = {
                'duplex': topo_neighbour.get('duplex'),
                'native_vlan': topo_neighbour.get('native_vlan'),
            }

            self.nv_objects.add_host(
                host=host,
                emblem=self.emblems.host_node,
            )
            self.nv_objects.add_host(
                host=neighbour_host,
                display_name=neighbour_name,
                emblem=self.emblems.host_node,
            )
            self.nv_objects.add_interface(
                host=str(host),
                service=str(local_port),
                metadata=metadata,
                display_name=str(raw_local_port) if self.display_ports else str(local_port),
                item=str(local_port),
                emblem=self.emblems.service_node,
            )
            self.nv_objects.add_interface(
                host=str(neighbour_host),
                service=str(neighbour_port),
                display_name=str(raw_neighbour_port) if self.display_ports else str(neighbour_port),
                item=str(neighbour_port),
                emblem=self.emblems.service_node,
            )
            self.nv_connections.add_connection(
                left=str(host),
                right=f'{local_port}@{host}',
            )
            self.nv_connections.add_connection(
                left=str(neighbour_host),
                right=f'{neighbour_port}@{neighbour_host}',
            )
            self.nv_connections.add_connection(
                left=f'{local_port}@{host}',
                right=f'{neighbour_port}@{neighbour_host}',
            )

    def adjust_raw_neighbour(self, raw_neighbour: str) -> str | None:
        """
        Checks if neighbour should be dropped or adjusted via regex.
        The next request for the same neighbour will be served from cache.
        Args:
            raw_neighbour: the neighbour name to check/adjust

        Returns:
            the adjusted neighbour name or None
        """
        try:
            return self.neighbour_to_host[raw_neighbour]
        except KeyError:
            pass

        if raw_neighbour in self.drop_neighbours:
            LOGGER.info(f'{self.topology} drop in {TomlSections.L2_DROP_NEIGHBOURS}: {raw_neighbour}')
            self.neighbour_to_host[raw_neighbour] = None
            return None

        adjusted_neighbour = raw_neighbour
        if self.neighbour_replace_regex:
            for re_str, replace_str in self.neighbour_replace_regex:
                re_neighbour = re_sub(re_str, replace_str, adjusted_neighbour)
                if not re_neighbour:
                    LOGGER.info(
                        f'{self.topology} removed by {TomlSections.L2_NEIGHBOUR_REPLACE_REGEX}: '
                        f'(|{adjusted_neighbour}|, |{re_str}|, |{replace_str}|)'
                    )
                    self.neighbour_to_host[raw_neighbour] = None
                    return None

                if re_neighbour != adjusted_neighbour:
                    LOGGER.info(
                        f'{self.topology} changed by {TomlSections.L2_NEIGHBOUR_REPLACE_REGEX}: '
                        f'|{adjusted_neighbour}| to |{re_neighbour}|'
                    )
                    adjusted_neighbour = re_neighbour

        self.neighbour_to_host[raw_neighbour] = adjusted_neighbour
        return adjusted_neighbour


class TopologyL3(Topology):
    def __init__(
            self,
            display_devices: bool,
            emblems: Emblems,
            host_cache: HostCache,
            ignore_hosts: Sequence[str],
            ignore_ips: Sequence[ip_network],
            ignore_wildcard: Sequence[Wildcard],
            include_hosts: bool,
            include_loopback: bool,
            replace: Mapping[str, str],
            skip_cidr_0: bool,
            skip_cidr_32_128: bool,
            skip_if: bool,
            skip_ip: bool,
            skip_public: bool,
            summarize: Sequence[ip_network],
            version: int
    ):
        super().__init__(
            emblems=emblems,
            host_cache=host_cache,
            topology=f'[L3 IPv{version}]',
        )
        self.diplay_devices = display_devices
        self.ignore_hosts: Sequence[str] = ignore_hosts
        self.ignore_ips: Sequence[ip_network] = ignore_ips
        self.ignore_wildcard: Sequence[Wildcard] = ignore_wildcard
        self.include_hosts: bool = include_hosts
        self.replace: Mapping[str, str] = replace
        self.show_loopback: bool = include_loopback
        self.skip_cidr_0: bool = skip_cidr_0
        self.skip_cidr_32_128: bool = skip_cidr_32_128
        self.skip_if: bool = skip_if
        self.skip_ip: bool = skip_ip
        self.skip_public: bool = skip_public
        self.summarize: Sequence[ip_network] = summarize
        self.version = version

    def create(self):
        match self.version:
            case IPVersion.IPv4:
                host_list: Sequence[str] = (
                    self.host_cache.query_hosts_by_filter(
                        HostFilter.LABELS, HostLabels.L3V4_ROUTER, LiveStatusOperator.EQUAL
                    )
                )

                if self.include_hosts:
                    host_list += self.host_cache.query_hosts_by_filter(
                        HostFilter.LABELS, HostLabels.L3V4_HOSTS, LiveStatusOperator.EQUAL
                    )

            case IPVersion.IPv6:
                host_list: Sequence[str] = self.host_cache.query_hosts_by_filter(
                    HostFilter.LABELS, HostLabels.L3V6_ROUTER, LiveStatusOperator.EQUAL
                )

                if self.include_hosts:
                    host_list += self.host_cache.query_hosts_by_filter(
                        HostFilter.LABELS, HostLabels.L3V6_HOSTS, LiveStatusOperator.EQUAL
                    )

            case _:
                host_list = []

        # check against filter list (host labels/attributes include/exclude)
        pre_filter_len = len(host_list)
        host_list = [host for host in host_list if self.host_cache.is_host_allowed(host)]
        LOGGER.info(f'{self.topology} # hosts allowed: {len(host_list)}/{pre_filter_len}')

        LOGGER.debug(f'{self.topology} host(s) to work on: {host_list}')
        if not host_list:
            LOGGER.error(
                f'{self.topology} No (routing capable) host found. Check if "inv_ip_addresses.mkp" '
                'added/enabled and inventory and host label discovery has run.'
            )
            return

        LOGGER.debug(f'{self.topology} ignore hosts: {self.ignore_hosts}')
        for raw_host in host_list:
            host = raw_host
            if host in self.ignore_hosts:
                LOGGER.info(f'{self.topology} host ignored in {TomlSections.L3_IGNORE_HOSTS}: {host}')
                continue
            if not (inv_ip_addresses := self.host_cache.get_data(
                    host=host, item=CacheItems.inventory, path=InvPaths.L3)
            ):
                LOGGER.warning(f'{self.topology} no IP address inventory found for host: {host}')
                continue

            self.nv_objects.add_host(host=host)
            for inv_ip_address in inv_ip_addresses:
                emblem = self.emblems.ip_network
                try:
                    device = inv_ip_address[L3InvColumns.DEVICE]
                    interface_address: ip_interface = ip_interface(
                        f'{inv_ip_address[L3InvColumns.ADDRESS]}/{inv_ip_address[L3InvColumns.CIDR]}'
                    )
                except KeyError as e:
                    LOGGER.warning(f'{self.topology} drop IP missing data: {host}, {inv_ip_address}, {e}.')
                    continue

                # skip entries without prefix-length/netmask
                if self.skip_cidr_0 and interface_address.network.prefixlen == 0:
                    LOGGER.info(f'{self.topology} drop IP with CIDR "0": {host}, {interface_address.compressed}')
                    continue

                if interface_address.version != self.version:
                    LOGGER.info(
                        f'{self.topology} drop IP non IPv{self.version} version: {host}, {interface_address.compressed}'
                    )
                    continue

                if not self.show_loopback and interface_address.is_loopback:
                    LOGGER.info(f'{self.topology} drop IP is loopback: {host},  {interface_address.ip.compressed}')
                    continue

                if self.skip_public and not interface_address.is_private:
                    LOGGER.info(f'{self.topology} drop IP is public: {host},  {interface_address.ip.compressed}')
                    continue

                if interface_address.is_link_local and interface_address.version == 6:
                    LOGGER.info(f'{self.topology} drop IP is link-local: {host}, {interface_address.ip.compressed}')
                    continue

                # drop host addresses  /32 or /128 -> one IP ine network
                if self.skip_cidr_32_128 and interface_address.network.num_addresses == 1:
                    LOGGER.info(f'{self.topology} drop IP with CIDR (32 or /128): {host}, {interface_address.compressed}')
                    continue

                if self.is_ignore_ip(interface_address.ip.compressed):
                    LOGGER.info(f'{self.topology} drop IP in {TomlSections.L3_IGNORE_IP}: {host}, {interface_address.compressed}')
                    continue

                if self.is_ignore_wildcard(interface_address.ip.compressed):
                    LOGGER.info(f'{self.topology} drop IP in {TomlSections.L3V4_IGNORE_WILDCARD}: {host}, {interface_address.compressed}')
                    continue

                if network := self.get_network_summary(raw_ip_address=interface_address.ip.compressed):
                    emblem = self.emblems.l3_summarize
                    LOGGER.info(
                        f'{self.topology} Summarized IP in {TomlSections.L3_SUMMARIZE}: {interface_address.compressed} -> {network}'
                    )
                else:
                    network = f'{interface_address.network.compressed}'

                if network in self.replace.keys():
                    LOGGER.info(f'{self.topology} Replaced network in {TomlSections.L3_REPLACE_NETWORKS}: {network} -> {self.replace[network]}')
                    network = self.replace[network]
                    emblem = self.emblems.l3_replace

                self.nv_objects.add_ip_network(network=network, emblem=emblem)

                item = None
                if not self.skip_if:
                    item = self.get_service_by_interface(host=host, interface=device)

                if self.skip_if is True and self.skip_ip is True:
                    self.nv_connections.add_connection(left=host, right=network)
                elif self.skip_if is True and self.skip_ip is False:
                    self.nv_objects.add_ip_address(
                        host=host,
                        interface=None,
                        raw_ip_address=interface_address.ip.compressed,
                        emblem=self.emblems.ip_address,
                    )
                    self.nv_objects.add_tooltip_quickinfo(
                        f'{interface_address.ip.compressed}@{host}', 'Interface', device
                    )
                    self.nv_connections.add_connection(left=f'{host}', right=f'{interface_address.ip.compressed}@{host}')
                    self.nv_connections.add_connection(left=network, right=f'{interface_address.ip.compressed}@{host}')
                elif self.skip_if is False and self.skip_ip is True:
                    self.nv_objects.add_interface(
                        display_name=device if self.diplay_devices else item,
                        host=host,
                        item=item,
                        service=device,
                    )
                    self.nv_objects.add_tooltip_quickinfo(
                        f'{device}@{host}', 'IP-address', interface_address.ip.compressed
                    )
                    self.nv_connections.add_connection(left=f'{host}', right=f'{device}@{host}')
                    self.nv_connections.add_connection(left=network, right=f'{device}@{host}')
                else:
                    self.nv_objects.add_ip_address(
                        host=host,
                        interface=device,
                        raw_ip_address=interface_address.ip.compressed,
                        emblem=self.emblems.ip_address,
                    )
                    self.nv_objects.add_interface(
                        display_name=device if self.diplay_devices else item,
                        host=host,
                        item=item,
                        service=device,
                    )
                    self.nv_connections.add_connection(
                        left=host, right=f'{device}@{host}')
                    self.nv_connections.add_connection(
                        left=f'{device}@{host}',
                        right=f'{interface_address.ip.compressed}@{device}@{host}',
                    )
                    self.nv_connections.add_connection(
                        left=network, right=f'{interface_address.ip.compressed}@{device}@{host}',
                    )

    def get_network_summary(self, raw_ip_address: str) -> str | None:
        for network in self.summarize:
            try:
                if ip_network(raw_ip_address).subnet_of(network):
                    LOGGER.debug(f'{self.topology} IP address is in subnet: {raw_ip_address} -> ({network})')
                    return network.compressed
            except TypeError:
                pass
        return None

    def is_ignore_ip(self, raw_ip_address: str) -> bool:
        for ip in self.ignore_ips:
            try:
                if ip_network(raw_ip_address).subnet_of(ip):
                    LOGGER.debug(f'{self.topology} IP address is in ignore list: {raw_ip_address} -> ({ip})')
                    return True
            except TypeError:
                continue
        return False

    def is_ignore_wildcard(self, raw_ip_address: str) -> bool:
        int_ip_address = int(ip_address(raw_ip_address))
        for wildcard in self.ignore_wildcard:
            if int_ip_address & wildcard.int_wildcard == wildcard.bit_pattern:
                LOGGER.debug(
                    f'{self.topology} IP address matches ignore wildcard list: '
                    f'{raw_ip_address} -> ({wildcard.ip_address}/{wildcard.wildcard})'
                )
                return True
        return False


def map_speed_to_thickness(speed_to_map: int, speed_map: Sequence[Thickness]) -> int:
    thickness: int = 1  # use in case of empty MAP_SPEED_TO_THICKNESS
    for speed, thickness in speed_map:
        if speed_to_map <= speed:
            break
    return thickness


def add_tooltip_html(
        metadata: Dict,
        type_: str,
        left: str,
        left_value: str,
        right: str,
        right_value: str,
) -> Dict:
    css_class = 'class=mismatch'
    if metadata.get('tooltip') is None:
        metadata['tooltip'] = {}
    if metadata['tooltip'].get('html') is None:
        css_style = (
            '<style>'
            'div.mismatch {'
            'background-color: rgba(70,70,70,.1);'
            'border-radius: 5px;'
            'padding: 5px;'
            '}'
            'p.mismatch {'
            'text-align: center;'
            '}'
            'table.mismatch {'
            'text-align: left;'
            'border-radius: 5px;'
            '}'
            'tr.mismatch {'
            '}'
            'tr.mismatch:nth-child(even) {'
            'background-color: rgba(100,100,100,.3);'

            '}'
            'tr.mismatch:nth-child(odd) {'
            'background-color: rgba(100,100,100,.2);'

            '}'
            'th.mismatch {'
            'background-color: rgba(120,120,120,.3);'
            'padding: 5px;'  # inside the element
            # 'margin: 5px;'  # outside of the lement
            '}'
            'td.mismatch {'
            'padding: 5px;'
            '}'
            '</style>'
        )
        header = (
            '<thead>'
            '<tr>'
            f'<th {css_class}>Node</th>'
            f'<th {css_class}>{left}</th>'
            f'<th {css_class}>{right}</th>'
            '</tr>'
            '</thead>'
        )
        metadata['tooltip']['html'] = (
            f'{css_style}'
            f'<div {css_class}>'
            f'<p {css_class}>WARNING: Mismatch found!</p>'
            f'<table {css_class}>'
            f'{header}'
            '<tbody>'
        )

    metadata['tooltip']['html'] += (
        f'<tr {css_class}>'
        f'<td {css_class}>{type_}</td>'
        f'<td {css_class}>{left_value}</td>'
        f'<td {css_class}>{right_value}</td>'
        '</tr>'
    )

    return metadata


def close_tooltip_html(metadata: Dict) -> Dict:
    if metadata.get('tooltip', {}).get('html'):
        metadata['tooltip']['html'] += '</table></div>'
    return metadata
