#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 202-01-03
# File  : nvdct/lib/adjust_toml.py


from pathlib import Path
from re import findall as re_findall, sub as re_sub
from sys import exit as sys_exit

from lib.constants import (
    API_PORT_DEFAULT,
    Backends,
    EmblemNames,
    EmblemValues,
    ExitCodes,
    IncludeExclude,
    L2Case,
    L2IgnoreMismatch,
    L2RemoveDomain,
    Layers,
    LogLevels,
    TomlSections,
    TomlSettings,
)


def update_config(toml_file: str):
    fix_options = {
        'DROP_HOSTS': TomlSections.L2_DROP_NEIGHBOURS,
        'HOST_MAP': TomlSections.L2_NEIGHBOUR_TO_HOST_MAP,
        'L2_DROP_HOSTS': TomlSections.L2_DROP_NEIGHBOURS,
        'L2_HOST_MAP': TomlSections.L2_NEIGHBOUR_TO_HOST_MAP,
        'L3V4_IGNORE_HOSTS': TomlSections.L3_IGNORE_HOSTS,
        'L3V4_IGNORE_IP': TomlSections.L3_IGNORE_IP,
        'L3V4_IRNORE_WILDCARD': TomlSections.L3V4_IGNORE_WILDCARD,
        'L3V4_REPLACE': TomlSections.L3_REPLACE_NETWORKS,
        'L3V3_REPLACE': TomlSections.L3_REPLACE_NETWORKS,
        'L3_REPLACE': TomlSections.L3_REPLACE_NETWORKS,
        'L3V4_SUMMARIZE': TomlSections.L3_SUMMARIZE,
        'SEED_DEVICES': TomlSections.L2_SEED_DEVICES,
        'SITES': TomlSections.FILTER_BY_SITE,
        'CUSTOMERS': TomlSections.FILTER_BY_CUSTOMER,
        'icon_missinc': EmblemValues.ICON_ALERT_UNREACHABLE,
        'icon_missing': EmblemValues.ICON_ALERT_UNREACHABLE,
        'l3v4_replace': EmblemNames.L3_REPLACE,
        'l3v4_summarize': EmblemNames.L3_SUMMARIZE,
        'keep_domain = true': f'{TomlSettings.L2_REMOVE_DOMAIN} = "{L2RemoveDomain.OFF}"',
        'keep_domain = false': f'{TomlSettings.L2_REMOVE_DOMAIN} = "{L2RemoveDomain.ON}"',
        'case': TomlSettings.L2_CASE,
        'prefix': TomlSettings.L2_PREFIX,
        'remove_domain': TomlSettings.L2_REMOVE_DOMAIN,
        'keep': TomlSettings.KEEP_MAX_TOPOLOGIES,
        'min_age': TomlSettings.MIN_TOPOLOGY_AGE,
        'display_l2_neighbours': TomlSettings.L2_DISPLAY_NEIGHBOURS,
        'include_l3_hosts': TomlSettings.L3_INCLUDE_HOSTS,
        'include_l3_loopback': TomlSettings.L3_INCLUDE_LOOPBACK,
        'skip_l3_cidr_0': TomlSettings.L3_SKIP_CIDR_0,
        'skip_l3_cidr_32_128': TomlSettings.L3_SKIP_CIDR_32_128,
        'skip_l3_if': TomlSettings.L3_SKIP_IF,
        'skip_l3_ip': TomlSettings.L3_SKIP_IP,
        'skip_l3_public': TomlSettings.L3_SKIP_PUBLIC,
    }

    fix_params = {
        f'# {TomlSettings.L2_CASE} = "{L2Case.LOWER}" | "{L2Case.UPPER}"\n': f'# {TomlSettings.L2_CASE} = "{L2Case.LOWER}" | "{L2Case.UPPER}" | {L2Case.INSENSITIVE} | "{L2Case.AUTO}" | "{L2Case.OFF}"\n',
        f'# {TomlSettings.L2_CASE} = "{L2Case.LOWER}\n"' : f'# {TomlSettings.L2_CASE} = "{L2Case.LOWER}" | "{L2Case.UPPER}" | {L2Case.INSENSITIVE} | "{L2Case.AUTO}" | "{L2Case.OFF}"\n',
        f'# {TomlSettings.L2_CASE} = "{L2Case.UPPER}\n"': f'# {TomlSettings.L2_CASE} = "{L2Case.LOWER}" | "{L2Case.UPPER}" | {L2Case.INSENSITIVE} | "{L2Case.AUTO}" | "{L2Case.OFF}\n',
        f'# {TomlSettings.L2_REMOVE_DOMAIN} = false | true\n': f'# {TomlSettings.L2_REMOVE_DOMAIN} = "{L2RemoveDomain.OFF}" | "{L2RemoveDomain.ON}" | "{L2RemoveDomain.AUTO}"\n',
        f'# {TomlSettings.L2_REMOVE_DOMAIN} = false\n': f'# {TomlSettings.L2_REMOVE_DOMAIN} = "{L2RemoveDomain.OFF}" | "{L2RemoveDomain.ON}" | "{L2RemoveDomain.AUTO}"\n',
        f'# {TomlSettings.L2_REMOVE_DOMAIN} = true\n': f'# {TomlSettings.L2_REMOVE_DOMAIN} = "{L2RemoveDomain.OFF}" | "{L2RemoveDomain.ON}" | "{L2RemoveDomain.AUTO}"\n',
        f'{TomlSettings.L2_REMOVE_DOMAIN} = false\n': f'{TomlSettings.L2_REMOVE_DOMAIN} = "{L2RemoveDomain.OFF}"\n',
        f'{TomlSettings.L2_REMOVE_DOMAIN} = true\n': f'{TomlSettings.L2_REMOVE_DOMAIN} = "{L2RemoveDomain.ON}"\n',
    }

    old_options = {
        'CUSTOM_LAYERS': f'Can be removed (no longer supported).',
        'FILESYSTEM': f'Use {Backends.MULTISITE} instead.',
        'debug': f'Use "{TomlSettings.LOG_LEVEL} = {LogLevels.DEBUG}" instead.',
        'lowercase': f'Use "{TomlSettings.L2_CASE} = {L2Case.LOWER}" instead.',
        'uppercase': f'Use "{TomlSettings.L2_CASE} = {L2Case.UPPER}" instead.',
        'CUSTOM': f'Is no loger supported. remove it from layers please.'
    }
    # sorted in reverse
    missing_settings = {
        TomlSettings.TIME_FORMAT: '"%Y-%m-%dT%H:%M:%S.%m"',
        TomlSettings.QUIET: 'false | true',
        TomlSettings.PRE_FETCH: 'false | true',
        TomlSettings.OUTPUT_DIRECTORY: '"nvdct"  # remove to get date formated directory',
        TomlSettings.MIN_TOPOLOGY_AGE: '1',
        TomlSettings.LOG_TO_STDOUT: 'false | true',
        TomlSettings.LOG_LEVEL: f'"{LogLevels.WARNING}" | "{LogLevels.DEBUG}" | "{LogLevels.INFO}" | "{LogLevels.ERROR}" | "{LogLevels.FATAL}" | {LogLevels.CRITICAL}" | "{LogLevels.OFF}"',
        TomlSettings.LOG_FILE: '"~/var/log/nvdct.log"',
        TomlSettings.LAYERS: f'["{Layers.CDP}", "{Layers.LLDP}", "{Layers.L3V4}", "{Layers.STATIC}"]',
        TomlSettings.L3_SKIP_PUBLIC: 'false | true',
        TomlSettings.L3_SKIP_IP: 'false | true',
        TomlSettings.L3_SKIP_IF: 'false | true',
        TomlSettings.L3_SKIP_CIDR_32_128: 'false | true',
        TomlSettings.L3_SKIP_CIDR_0: 'false | true',
        TomlSettings.L3_INCLUDE_LOOPBACK: 'false | true  # most likely dropped from inventory (SNMP) before',
        TomlSettings.L3_INCLUDE_HOSTS: 'false | true',
        TomlSettings.L3_DISPLAY_DEVICES: 'false | true',
        TomlSettings.L2_SKIP_EXTERNAL: 'false | true',
        TomlSettings.L2_REMOVE_DOMAIN: f'"{L2RemoveDomain.OFF}" | "{L2RemoveDomain.ON}" | "{L2RemoveDomain.AUTO}"',
        TomlSettings.L2_PREFIX: '""',
        TomlSettings.L2_IGNORE_MISMATCH: f'["{L2IgnoreMismatch.DUPLEX}", "{L2IgnoreMismatch.SPEED}", "{L2IgnoreMismatch.VLAN}"]',
        TomlSettings.L2_DISPLAY_NEIGHBOURS: 'false | true',
        TomlSettings.L2_DISPLAY_PORTS: 'false | true',
        TomlSettings.L2_CASE: f'"{L2Case.OFF}" | "{L2Case.LOWER}" | "{L2Case.UPPER}" | {L2Case.INSENSITIVE} | "{L2Case.AUTO}"',
        TomlSettings.KEEP_MAX_TOPOLOGIES: '10',
        TomlSettings.FILTER_SITES: f'"{IncludeExclude.INCLUDE}" | "{IncludeExclude.EXCLUDE}"',
        TomlSettings.FILTER_CUSTOMERS: f'"{IncludeExclude.INCLUDE}" | "{IncludeExclude.EXCLUDE}"',
        TomlSettings.DONT_COMPARE: 'false | true',
        TomlSettings.DEFAULT: 'false | true',
        TomlSettings.BACKEND: f'"{Backends.MULTISITE}" | "{Backends.RESTAPI}" | "{Backends.LIVESTATUS}"',
        TomlSettings.API_PORT: API_PORT_DEFAULT,
    }

    l2_neighbour_to_host_map = r'''
# map inventory CDP/LLDP neighbour name to Checkmk host name
# [0-9-a-zA-Z\.\_\-]{1,253} -> host
    
# "inventory_neighbour1" = "cmk_host1"
# "inventory_neighbour2" = "cmk_host2"
# "inventory_neighbour3" = "cmk_host3"
    '''

    l2_neighbour_replace_regex = r'''
# modify CDP/LLDP neighbour name with regex before mapping to CMK host names

# "regex string to replace" = "string to replace with"
# "^(([0-9a-fA-F]){2}[:.-]?){5}([0-9a-fA-F]){2}$" = ""
# "\\([0-9a-zA-Z]+\\)$" = ""
# "^Meraki.*\\s-\\s" = ""    
    '''

    l3_replace_networks = r'''
# replace _network objects_ in L§ topologies (takes place after summarize)
# [0-9-a-zA-Z\.\_\-]{1,253} -> host

# "10.193.172.0/24" = "MPLS"
# "10.194.8.0/23" = "MPLS"
# "10.194.12.0/24" = "MPLS"
# "10.194.115.0/24" = "MPLS"
# "fc00::/7" = "Unique-local"    
    '''

    emblems = f'''
# can use misc icons from CMK or upload your own in the misc category
# for built-in icons use "icon_" as l2_prefix to the name from CMK
# max size 80x80px
# emblems will only be used for non CMK objects

# "{EmblemNames.HOST_NODE}" = "{EmblemValues.ICON_ALERT_UNREACHABLE}"
# "{EmblemNames.IP_ADDRESS}" = "{EmblemValues.IP_ADDRESS_80}"
# "{EmblemNames.IP_NETWORK}" = "{EmblemValues.IP_NETWORK_80}"
# "{EmblemNames.L3_REPLACE}" = "{EmblemValues.ICON_PLUGINS_CLOUD}"
# "{EmblemNames.L3_SUMMARIZE}" = "{EmblemValues.ICON_AGGREGATION}"
# "{EmblemNames.SERVICE_NODE}" = "{EmblemValues.ICON_ALERT_UNREACHABLE}"    
    '''

    map_speed_to_thickness = r'''
# must be sorted from slower to faster speed
# use only one/no entry to have all connections with the same thickness
# "bits per second" = thickness

# "2000000" = 1  # 2 mbit
# "5000000" = 2  # 5 mbit
# "1e7" = 3      # 10 mbit
# "51e7" = 4     # 51 mbit
"1e8" = 1  # 100 mbit
"1e9" = 3  # 1 gbit
"1e10" = 5 # 10 gbit    
    '''

    filter_by_folder = f'''
# "/folder1/subfolder1" = "{IncludeExclude.INCLUDE}" | "{IncludeExclude.EXCLUDE}"
# "/folder2/subfolder2" = "{IncludeExclude.INCLUDE}" | "{IncludeExclude.EXCLUDE}"    
'''

    filter_host_by_label = f'''
# "hostlabel1:value" = "{IncludeExclude.INCLUDE}" | "{IncludeExclude.EXCLUDE}"
# "hostlabel2:value" = "{IncludeExclude.INCLUDE}" | "{IncludeExclude.EXCLUDE}"
'''

    filter_host_by_tag = f'''
# "host_tag1:value" = "{IncludeExclude.INCLUDE}" | "{IncludeExclude.EXCLUDE}"
# "host_tag2:value" = "{IncludeExclude.INCLUDE}" | "{IncludeExclude.EXCLUDE}"
    '''

    missing_tables = {
        TomlSections.L2_NEIGHBOUR_TO_HOST_MAP: l2_neighbour_to_host_map,
        TomlSections.L2_NEIGHBOUR_REPLACE_REGEX: l2_neighbour_replace_regex,
        TomlSections.L3_REPLACE_NETWORKS: l3_replace_networks,
        TomlSections.EMBLEMS: emblems,
        TomlSections.MAP_SPEED_TO_THICKNESS: map_speed_to_thickness,
        TomlSections.FILTER_BY_FOLDER: filter_by_folder,
        TomlSections.FILTER_BY_HOST_LABEL: filter_host_by_label,
        TomlSections.FILTER_BY_HOST_TAG: filter_host_by_tag,
    }

    l2_seed_devices = r''' = [
    # list of CDP/LLDP seed devices (if empty, all CDP/LLDP devices will be used)
    # [0-9-a-zA-Z\.\_\-]{1,253} -> host

    # "CORE01",
    # "LOCATION01",
    # "LOCATION02",
]
    '''

    l2_drop_neighbours = r''' = [
    # drop CDP/LLDP neighbours names

    # "not advertised",
    # "a nother invalid name",
]
    '''

    l3_ignore_hosts = r''' = [
    # hosts will be ignored in L3 topologies
    # [0-9-a-zA-Z\.\_\-]{1,253} -> host
    
    # "host1",
    # "host2",
]
    '''

    l3_ignore_ip = r''' = [
    # drop IP address that matches ip/network

    # "192.168.100.231",
    # "192.168.100.0/16",
    # "192.168.150.0/255.255.255.0",
    # "fd00::1"
    # "fd00::/8"
]
    '''

    l3v4_ignore_wildcard = r''' = [
    # ignore IPs by wildcard
    # if comparing an ip address:
    # each 0 bit in the wildcard has to be exactly as in the pattern
    # each 1 bit in the wildcard will be ignored

    # [ pattern    ,  wildcard ]
    # ["172.17.0.1", "0.0.255.0"],  # ignore all IPs ending with 1 from 172.17.0.0/16
    # ["172.17.128.0", "0.0.127.3"],  # ignore all IPs ending with 0-3 from 172.17.128.0/17
    # ["172.17.128.3", "0.0.127.0"],  # ignore all IPs ending with 3 from 172.17.128.0/17
]
'''

    l3_summarize = r''' = [
    # IP _networks_ to summarize

    # "10.193.172.0/24",
    # "10.194.8.0/23",
    # "10.194.12.0/24",
    # "10.194.115.0/255.255.255.0",
    # "fd00::/8"
]
    '''

    protected_topologies = r''' = [
    # topologies will not be deleted by "--keep_max_topologies"

    # "2023-10-17T14:08:05.10",
    # "your_important_topology"
]
    '''

    static_connections = r''' = [
    # user defined static connections
    # [0-9-a-zA-Z\.\_\-]{1,253} -> host

    # valid entry formats
    # ["left_host", "left_service", "right_service", "right_host"],
    # connection: "left_host"<->"left_service"<->"right_service"<->"right_host"
    # ["left_host", "", "right_service", "right_host"],
    # connection: "left_host"<->"right_service"<->"right_host"
    # ["left_host", "left_service", "", "right_host"],
    # connection: "left_host"<->"left_service"<->"right_host"
    # ["left_host", "", "", "right_host"],
    # connection: "left_host"<->"right_host"
]
    '''

    filter_by_customer = r''' = [
    # list customers to include/exclude, use with option --filter-costumers INCLUDE/EXCLUDE
    # [0-9-a-zA-Z\.\_\-]{1,16} -> customer
    
    # "customer1",
    # "customer2",
    # "customer3"
]
    '''

    filter_by_site = r''' = [
    # list site to include/exclude, use with option --filter-sites INCLUDE/EXCLUDE
    # [0-9-a-zA-Z\.\_\-]{1,16} -> site
    
    # "site1",
    # "site2",
    # "site3",
]
    '''

    missing_arrays = {
        TomlSections.FILTER_BY_SITE: filter_by_site,
        TomlSections.FILTER_BY_CUSTOMER: filter_by_customer,
        TomlSections.STATIC_CONNECTIONS: static_connections,
        TomlSections.PROTECTED_TOPOLOGIES: protected_topologies,
        TomlSections.L3_SUMMARIZE: l3_summarize,
        TomlSections.L3V4_IGNORE_WILDCARD: l3v4_ignore_wildcard,
        TomlSections.L3_IGNORE_IP: l3_ignore_ip,
        TomlSections.L3_IGNORE_HOSTS: l3_ignore_hosts,
        TomlSections.L2_DROP_NEIGHBOURS: l2_drop_neighbours,
        TomlSections.L2_SEED_DEVICES: l2_seed_devices,
    }

    changed: bool = False
    org_file = Path(toml_file)
    if not org_file.exists():
        print(f'Config file {org_file.name} not found!')
        sys_exit(ExitCodes.FILE_NOT_FOUND)

    print(f'Checking file.: {org_file.name}')
    org_content: str = org_file.read_text()
    content: str = org_content
    for old, new in fix_options.items():
        re_pattern = f'\\b{old}\\b'
        count = len(re_findall(re_pattern, content))
        if count > 0:
            changed = True
            content = re_sub(re_pattern, new, content)
            print(f'Found value...: "{old}" {count} times, replaced by "{new}"')

    for old, new in fix_params.items():
        if old in content:
            changed = True
            content = content.replace(old, new)
            new_line = '\n'
            print(f'Found value...: "{old.replace(new_line, "")}", replaced by "{new.replace(new_line, "")}"')

    for missing_setting, value in missing_settings.items():
        re_pattern = f'\\b{missing_setting}\\b'
        count = len(re_findall(re_pattern, content))
        if count == 0:
            changed = True
            content = re_sub(
                f'\\[{TomlSections.SETTINGS}\\]\n',
                f'[{TomlSections.SETTINGS}]\n# {missing_setting} = {value}\n',
                content
            )
            print(f'Added option..: "# {missing_setting} = {value}"')

    for table, value in missing_tables.items():
        re_pattern = f'\\[{table}\\]\n'
        count = len(re_findall(re_pattern, content))
        if count == 0:
            changed = True
            content = re_sub(
                f'\\[{TomlSections.SETTINGS}\\]\n',
                f'[{table}]{value}\n[{TomlSections.SETTINGS}]\n',
                content
            )
            print(f'Added section.: "[{table}]"')

    for array, value in missing_arrays.items():
        re_pattern = f'\\b{array}\\b'
        count = len(re_findall(re_pattern, content))
        if count == 0:
            changed = True
            content = content.replace(']\n\n#', f']\n\n#\n{array}{value}\n#', 1)
            print(f'Added section.: "{array} = []"')
            pass

    for old, new in old_options.items():
        re_pattern = f'\\b{old}\\b'
        count = len(re_findall(re_pattern, org_content))
        if count > 0:
            print(f'Obsolete......: "{old}", {new}')

    if changed:
        backup_file = Path(f'{toml_file}.backup')
        if not backup_file.exists():
            org_file.rename(backup_file)
            print(f'Renamed TOML..: {backup_file.name}')
            new_file = Path(toml_file)
            new_file.open('w').write(content)
            print(f'Written fixed.: {new_file.name}')
        else:
            print(
                f'Can not create backup file {backup_file.name}, file exists. Aborting!\n'
                f'Nothing has changed.'
            )
    else:
        print('Finished......: Nothing found to fix.')
