#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-10-12
# File  : nvdct/lib/utils.py

from ast import literal_eval
from collections.abc import Mapping, MutableSequence, Sequence
from json import dumps, loads
from logging import disable as log_off, Formatter, getLogger, StreamHandler
from logging.handlers import RotatingFileHandler
from pathlib import Path
from re import match as re_match
from socket import socket, AF_UNIX, AF_INET, SOCK_STREAM, SHUT_WR
from sys import stdout, exit as sys_exit
from time import time as now_time
from tomllib import loads as toml_loads, TOMLDecodeError
from typing import Dict, List, TextIO

from lib.constants import (
    CMK_SITE_CONF,
    CONFIG_FILE,
    DATAPATH,
    ExitCodes,
    LOGGER,
    OMD_ROOT,
)


def get_local_cmk_version() -> str:
    return Path(f'{OMD_ROOT}/version').readlink().name


def get_local_cmk_api_port() -> int | None:
    data = get_data_from_toml(CMK_SITE_CONF)
    return data.get('CONFIG_APACHE_TCP_PORT')


def get_data_form_live_status(query: str) -> Dict | List | None:
    address = f'{OMD_ROOT}/tmp/run/live'
    # family = AF_INET if type(address) is tuple else AF_UNIX
    family = AF_INET if isinstance(address, tuple) else AF_UNIX
    sock = socket(family, SOCK_STREAM)
    sock.connect(address)
    sock.sendall(query.encode())
    sock.shutdown(SHUT_WR)
    chunks: MutableSequence = []
    while len(chunks) == 0 or chunks[-1] != "":
        chunks.append(sock.recv(4096).decode())
    sock.close()
    if len(chunks) > 0:
        reply = literal_eval("".join(chunks).strip())
        if reply != [[b'']]:
            return reply
    return None


def get_data_from_toml(file: str, check_only:bool = False) -> Dict| bool:
    data = {}
    toml_file = Path(file)
    if toml_file.exists():
        try:
            data = toml_loads(toml_file.read_text())
        except TOMLDecodeError as e:
            LOGGER.exception(
                f'Config file {toml_file} is not in valid TOML format! ({e}),'
                f' (see https://toml.io/en/)'
            )
            sys_exit(ExitCodes.BAD_TOML_FORMAT)
        if check_only:
            message: str = f'Could read/parse the data from {toml_file}'
            # will not be logged as logger si not initialized yet
            # LOGGER.info(message)
            print(message)
            sys_exit(ExitCodes.OK)
    else:
        LOGGER.error(f'\nConfig {file} not found. Falling back to default config file "{CONFIG_FILE}"')
    # will not be logged as logger ist not initialized.
    # LOGGER.info(f'Config file read: {file}')
    # LOGGER.debug(f'Data from config file: {data}')
    return data


def rm_tree(root: Path) -> None:
    # safety
    if not str(root).startswith(DATAPATH):
        LOGGER.warning(f"Bad path to remove, {str(root)}, don\'t delete it.")
        return
    for p in root.iterdir():
        if p.is_dir():
            rm_tree(p)
        else:
            p.unlink()
    root.rmdir()


def remove_old_data(keep: int, min_age: int, raw_path: str, protected: Sequence[str]) -> None:
    if not keep:
        return
    path: Path = Path(raw_path)
    default_topo = path.joinpath('default')
    directories = [str(directory) for directory in list(path.iterdir())]

    # keep default topo
    if str(default_topo) in directories:
        directories.remove(str(default_topo))
        keep -= 1
        if default_topo.is_symlink():
            try:
                directories.remove(str(path.joinpath(str(default_topo.readlink()))))
            except ValueError:
                pass

    # keep protected topologies
    for directory in protected:
        try:
            directories.remove(str(path.joinpath(directory)))
        except ValueError:
            pass
        else:
            LOGGER.info(f'Protected topology: {directory}, will not be deleted.')

    if len(directories) < keep < 1:
        return

    topo_by_age = {}

    for directory in directories:
        if Path(directory).is_dir():
            topo_by_age[Path(directory).stat().st_ctime] = directory

    topo_age: List = list(topo_by_age.keys())
    topo_age.sort(reverse=True)

    while len(topo_by_age) > keep:
        entry = topo_age.pop()
        if min_age * 86400 > now_time() - entry:
            LOGGER.info(
                f'Topology "{Path(topo_by_age[entry]).name}'
                f'" not older then {min_age} day(s). not deleted.'
            )
            return
        LOGGER.info(f'delete old topology: {topo_by_age[entry]}')
        rm_tree(Path(topo_by_age[entry]))
        topo_by_age.pop(entry)


def save_data_to_file(
        data: Mapping,
        path: str,
        file: str,
        make_default: bool,
        dont_compare: bool,
) -> None:
    """
    Save the data as json file.
    Args:
        data: the topology data
        path: the path inder DATATAPATH
        file: the file name to save the data in
        make_default: if True, create the symlink "default" with path as target
        dont_compare: if True, data will not be compared with default data
    Returns:
        None
    """
    if not Path(f'{DATAPATH}/default').exists():
        make_default = True
    elif Path(f'{DATAPATH}/default/{file}').exists() and not dont_compare:
        if is_equal_with_default(data, f'{DATAPATH}/default/{file}'):
            LOGGER.warning(f'Data identical to default. Not saved! Use "--dont-compare".')
            return

    path_file = f'{DATAPATH}/{path}/{file}'
    save_file = Path(f'{path_file}')
    save_file.parent.mkdir(exist_ok=True, parents=True)
    save_file.write_text(dumps(data))


    if make_default:
        Path(f'{DATAPATH}/default').unlink(missing_ok=True)
        Path(f'{DATAPATH}/default').symlink_to(target=Path(path), target_is_directory=True)


def is_list_of_str_equal(list1: List[str], list2: List[str]) -> bool:
    """
    Compares two list of strings. Before comparing the list will internal sorted.
    Args:
        list1:
        list2:

    Returns:
        True if both lists match
        False if they don't match
    """
    tmp_list1 = list1.copy()
    tmp_list2 = list2.copy()
    try:
        tmp_list1.sort()
        tmp_list2.sort()
    except TypeError:  # list of dict cant be sorted (?)
        pass
    return tmp_list1 == tmp_list2


def is_valid_hostname(host: str) -> bool:
    re_host_pattern = r'^[0-9a-z-A-Z\.\-\_]{1,253}$'
    if re_match(re_host_pattern, host):
        return True
    else:
        LOGGER.error(f'Invalid hostname found: |{host}|')
        return False


def is_valid_site_name(site: str) -> bool:
    re_host_pattern = r'^[0-9a-z-A-Z\.\-\_]{1,16}$'
    if re_match(re_host_pattern, site):
        return True
    else:
        LOGGER.error(f'Invalid site name found: |{site}|')
        return False


def is_valid_customer_name(customer: str) -> bool:
    re_host_pattern = r'^[0-9a-z-A-Z\.\-\_]{1,16}$'
    if re_match(re_host_pattern, customer):
        return True
    else:
        LOGGER.error(f'Invalid customer name found: |{customer}|')
        return False


def is_valid_output_directory(directory: str) -> bool:
    # 2024-12-11T17:35:08.12
    re_host_pattern = r'^[0-9a-z-A-Z\.\-\_\:]{1,30}$'
    if re_match(re_host_pattern, directory):
        return True
    else:
        LOGGER.error(f'Invalid output directory name found: {directory}')
        return False


def compare_dicts(dict1: Mapping, dict2: Mapping) -> bool:
    # check top level keys
    if not is_list_of_str_equal(list(dict1.keys()), list(dict2.keys())):
        LOGGER.debug('Top level dont match')
        LOGGER.debug(f'dict1: {list(dict1.keys())}')
        LOGGER.debug(f'dict1: {list(dict2.keys())}')
        return False

    LOGGER.debug('Top level matches')
    for key, value in dict1.items():
        type_ = type(value)
        if type_ == dict:
            LOGGER.debug(f'compare dict: {key}')
            if not compare_dicts(value, dict2[key]):
                return False
        elif type_ == list:
            if not is_list_of_str_equal(value, dict2[key]):
                LOGGER.debug(f'list1: {value}')
                LOGGER.debug(f'list2: {dict2[key]}')
                return False
        elif type_ in [str, int]:
            if not value == dict2[key]:
                LOGGER.debug('value dont match')
                LOGGER.debug(f'value1: {value}')
                LOGGER.debug(f'value2: {dict2[key]}')
                return False
        elif value is None:
            if not value == dict2[key]:
                LOGGER.debug('value dont match')
                LOGGER.debug(f'value1: {value}')
                LOGGER.debug(f'value2: {dict2[key]}')
                return False
        else:
            LOGGER.debug(f'Compare unknown type {type_}')
            LOGGER.debug(f'key: {key}')
            LOGGER.debug(f'value1: {value}')
            LOGGER.debug(f'value2: {dict2[key]}')
            return False

    return True


def is_equal_with_default(data: Mapping, file: str) -> bool:
    default_file = Path(file)
    if default_file.exists():
        LOGGER.info(f'compare data with {file}')
        default_data = loads(default_file.read_text())
        return compare_dicts(data, default_data)
    return False


def get_data_from_inventory(inventory: Dict[str, object], raw_path: str) -> List | Dict | None:
    # path: List[str] = ('Nodes,' + ',Nodes,'.join(raw_path.split(',')) + ',Table,Rows').split(',')
    split_path: List[str] = raw_path.split(',')
    path: MutableSequence[str] = []
    for entry in split_path[0:-2]:
        path += ['Nodes', entry]
    path += split_path[-2:]

    try:
        table = inventory.copy()
    except AttributeError:
        return None
    for m in path:
        try:
            table = table[m]
        except KeyError:
            LOGGER.info(f'Inventory table not found for: {path}')
            return None

    return table


def configure_logger(log_file: str, log_level: int, log_to_console: bool) -> None:
    if log_level < 0:
        log_off()
        return

    log = getLogger()
    log_formatter = Formatter(
        fmt='%(asctime)s :: %(levelname)s :: %(module)s :: %(funcName)s() :: %(lineno)s :: %(message)s',
    )

    log.setLevel(log_level)

    # create a new file > 5 mb size
    log_handler_file = RotatingFileHandler(
        filename=log_file,
        mode='a',
        maxBytes=5 * 1024 * 1024,
        backupCount=5,
        # encoding=None,
        # delay=0
    )

    log_handler_file.setFormatter(log_formatter)
    # log_handler_file.setLevel(INFO)
    log.addHandler(log_handler_file)

    if log_to_console:
        log_handler_console = StreamHandler(stdout)
        log_handler_console.setFormatter(log_formatter)
        # log_handler_console.setLevel(INFO)
        log.addHandler(log_handler_console)


class StdoutQuiet:
    def __init__(self, term: TextIO = stdout, quiet: bool = False):
        # super().__init__()
        self.term = term
        self._quiet = quiet
        self._org_stdout = stdout

    def write(self, text):
        if not self._quiet:
            self.term.write(text)
        else:
            pass

    def flush(self):
        self._org_stdout.flush()
