#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2

# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-10-08
# File  : nvdct/nvdct.py

# 2023-10-10: initial release
# 2023-10-16: added options  --keep-max  and --min-age
# 2023-10-17: removed -p, -c, -d (use the long options instead: --path-in-inventory, --data-source,
#             --inventory-columns)
#             changed option --keep-max to --keep
#             added option --check-user-data-only
#             added SEED_DEVICES to user data, option -s/--sed-devices is now optional
#             refactoring
# 2023-10-18: fixed --make-default not working
#             added protected topologies
#             changed get inventory data from file access to request via live status query
#             get interfaces from autochecks changed to live status query
# 2023-10-19: cleanup removed "all" unused functions
#             added option -u, --user-data-file to provide a user specific data file
# 2023-10-20: changed option -m/--make-default to -d/--default (-m is needed for -m/--merge
#             added -m/--merge option
# 2023-10-27: improved matching of port name from CDP/LLDP data to interface (item) name of service
#             improved internal data handling (HostCache)
# 2023-10-28: reworked handling of layers (see option -l/--layers)
#             added option -l/-layers {CDP,CUSTOM,LLDP,STATIC} [{CDP,CUSTOM,LLDP,STATIC} ...]
#             added section CUSTOM_LAYERS to user data file
#             removed option -m/--merge, included in option -l/--layers
#             removed option --lldp, included in option -l/--layers
#             removed option --data-source, now handled in CUSTOM_LAYERS
#             removed option --inventory-columns, now handled in CUSTOM_LAYERS
#             removed option --path-in-inventory, now handled in CUSTOM_LAYERS
#             removed lower limits for --keep and --min-age
# 2023-1ß-29: fixed missing path from custom layer in host prefetch
# 2023-11-16: added option -b/--backend [LIVESTATUS, FILESYSTEM],
#             LIVESTATUS is the default for performance reasons -> for now only local site
#             FILESYSTEM fetches the data directly form the inventory files -> use in distributed
#                        environments
# 2023-11-28: implemented MULTISITE as backend, via livestatus -> use in distributed environments,
#             may become the default
# 2023-11-30: added RESTAPI backend --> this is kind of slow (20 seconds compared to 0.5
#             for MULTISITE)
# 2023-12-06: do not add devices from static connections to seed devices --> be more transparent
#             added prefix option -> needs testing (THX to andreas doehler for the suggestion)
# 2023-12-21: adjusted for streamlined CDP/LLDP inventory format,
#             needs at least CDP 0.7.1-20231221 and LLDP 0.9.2-2023122
#             changed default backend from LIVESTATUS to MULTISITE
#             added logger (option --log-level)
#             removed option --debug
# 2023-12-22: added options --log-file and --log-to-stdout
# 2024-01-02: added new output format (option --new-format)
#             fixed LOGGER.warn --> LOGGER.warning
#             improved detection of missing hosts/interfaces
# 2024-01-03: fixed query in MULTISITE.exist
#             added option --quiet
#             renamed from create_topology_data to nvdct
#             added layer3 IPv4 topology (experimental)
# 2024-01-04: added SETTINGS section to nvdct.toml file -> defaults -> overridden by
#             toml -> overridden by cli
#             only seed devices are additive (toml + cli)
#             added L3v4 toml options: L3V4_IGNORE_HOSTS, L3V4_SUMMARIZE, L3V3_REPLACE
#             moved from ~/local/bin/topology_data to ~/local/bin/nvdct
# 2024-01-28: fixed -b MULTISITE if not in distributed environment (falling back to local)
#             fixed crash in create_l3v4_topology if no host was found
# 2024-01-29: added new output path for cmk 2.3
#             changed don't output cmk 2.2 format if --new-format set
# 2024-02-25: added images for L3 Topo (IP-Address and IP-Network)
#             icon source includes: https://www.flaticon.com/free-icons
#             added option to drop neighbours by regex
#             fixed no inventory in L3v4
# 2024-03-14: added metadata for connection (thickness, color)
#             thickness can be configured via toml (MAP_SPEED_TO_THICKNESS)
#             color 'red' is used for speed mismatch
#             made node emblem configurable (for ip-address and ip-network) via toml (EMBLEMS)
# 2024-03-16: moved args, utils, classes to lib
#             split classes in to backends, settings and classes
# 2024-03-17: changed speed mismatch from native speed to operational speed
#             added duplex mismatch ('yellow')
#             added native vlan mismatch ('orange')
# 2024-03-25: added --pre-fetch, this will speed up the RESTAPI backend performance
#             min inv_cdp_cache version: 0.7.1-20240320 -> host label nvdct/has_cdp_neighbours
#             min inv_lldp_cache version: 0.9.3-20240320 -> host label nvdct/has_lldp_neighbours
# 2024-03-27: added option --api-port, defaults to 80
# 2024-03-28: changed restapi get_interface_data to use one call to fetch all data
# 2024-03-29: added L3V4_IGNORE_IP and L3V4_IGNORE_WILDCARD to toml file
# 2024-03-31: fixed: L3V3_REPLACE -> L3V4_REPLACE
#             made input from toml more bullet proof
# 2024-04-03: change HostCacheMultiSite to be child of HostCacheLiveStatus
# 2024-04-06: added node icon warning
# 2024-04-07: added css for connection
# 2024-04-13: added html tooltip for connection issues
# 2024-04-15: added emblems for l3v4_summary and l3v4_replace
# 2024-04-16: fixed html tooltip for speed mismatch (Speed: 100 MBit/s (expected: 1 GBit/s)WARN)
#             added option --skip-l3-if
# 2024-04-17: automatically detect local CMK version -> no need for --new-format option any more
#             automatically detect local CONFIG_APACHE_TCP_PORT -> no need for --api-port option any more  # noqa: E501
# 2024-04-18: added option --skip-l3-ip
#             added tooltip for --skip-l3-if and --skip-l3-ip
# 2024-04-30: automatically create @default if it doesn't exist
# 2024-05-07: fixed handling of ignore wildcard in L3v4 topology
# 2024-05-17: added support for Extreme interfaces in LLDP without stack id (1 -> 1:1)
# 2024-05-18: added support for NXOS interfaces in LLDP in short from (Eth1/1 -> Ethernet1/1)
#             added IP-Address/IP-Network as quickinfo
# 2024-05-18: fixed crash non empty neighbour port (ThX to andreas doehler)
# 2024-06-05: fixed interface index padding
# 2024-06-06: added interface detection alias+index
#             added interface detection description + index
# 2024-06-09: moved topology helpers to lib/topologies.py
#             moved (default) config file(s) to ./conf/
# 2024-06-14: added debug code for bad IPv4 address data
# 2024-06-17: fixed bad IPv4 address data (just drop it)
# 2024-09-23: INCOMPATIBLE: replaced options --lowercase/--uppercase with --case LOWER|UPPER
#             changed version output from settings to argparse action
#             INCOMPATIBLE: removed backend FILESYSTEM -> will fallback to MULTISITE
#             INCOMPATIBLE: removed support for CMK2.2.x file format (removed option --new-format)
# 2024-09-24: added site filter for multisite deployments (MULTISITE only), option --filter-sites
#               and SITES section in toml file
#             added customer filter for MSP deployments (MULTISITE only), option --filter-customers
#               and section CUSTOMERS in toml file
# 2024-11-16: added better logging for missing L2 data
# 2024-11-17: added L2_NEIGHBOUR_REPLACE_REGEX (ThX to Frankb@checkmk forum for the base idea)
#             INCOMPATIBLE: removed DROP_HOST_REGEX -> use L2_NEIGHBOUR_REPLACE_REGEX instead
#             INCOMPATIBLE: changed section names in TOML file to better distinguish between L2 and L3v4
#               HOST_MAP -> L2_HOST_MAP
#               DROP_HOSTS -> L2_DROP_HOSTS
#               SEED_DEVICES -> L2_SEED_DEVICES
#             INCOMPATIBLE_ removed option -s, --seed-devices from CLI -> use TOML section L2_SEED_DEVICES instead
#             INCOMPATIBLE: changed the option keep-domain to remove-domain -> don't mess with neighbor names by default
# 2024-12-08: INCOMPATIBLE: changed hostlabel for L3v4 topology to nvdct/l3v4_topology
#                           needs at least inv_ip_address inv_ip_address-0.0.5-20241209.mkp
# 2024-12-09: added option --include-l3-hosts
#             added site filter for RESTAPI backend
#             enabled customer filter for MULTISITE backend
# 2024-12-10: refactoring: moved topology code to topologies, removed all global variables, created main() function
# 2024-12-11: INCOMPATIBLE: changed default layers to None -> use the CLI option -l CDP or the configfile instead
#             INCOMPATIBLE: reworked static topology -> can now be used for each service, host/service name has to be
#                           exactly like in CMK. See ~/local/bin/nvdct/conf/nfdct.toml
#             moved string constants to lib/constants.py
#
#             minimum ip address inventory plugin versions
#               - inv_lnx_if_ip-0.0.4-20241210.mkp
#               - inv_ip_address-0.0.6-20241210.mkp
#               - inv_win_if_ip-0.0.3-20241210.mkp
# 2024-12-20: fixed typo in TOML L3v4 -> "L3v4" (ThX to BH2005@checkmk_forum)
#             fixed crash in topologies (devices_to_go.remove(device) ValueError if device not in list) (ThX to BH2005)
# 2024-12-23: streamlined L3v4 in preparation for L3v6 topology
#             INCOMPATIBLE: changes in TOML:
#                 L3V4_IGNORE_HOSTS    -> L3_IGNORE_HOSTS
#                 L3V4_IGNORE_IP       -> L3_IGNORE_IP
#                 L3V4_SUMMARIZE       -> L3_SUMMARIZE
#                 L3V4_REPLACE         -> L3_REPLACE
#                 L3V4_IRNORE_WILDCARD -> L3V4_IGNORE_WILDCARD  # Typo
#                 [EMBLEMS]
#                 l3v4_replace         -> l3_replace
#                 l3v4_summarize       -> l3_summarize
# 2024-12-25: fixed "--dont-compare", data will only be saved if the are different from the default
#             changed: is seed devices is not configured use all CDP/LLDP devices (by host label)
# 2024-12-26: INCOMPATIBLE: renamed L2_DROP_HOSTS -> L2_DROP_NEIGHBOURS
#             added option --display-l2-neighbours
# 2024-12-27: added options
#               --adjust-toml
#               --include-l3-loopback
#               --skip-l3-cidr-0
#               --skip-l3-cidr-32-128
#               --skip-l3-public
#             fixed: keep default topology
#             fixed: cleanup -> remove the oldest topologies not the newest
#             INCOMPATIBLE: removed: CUSTOM_LAYERS
#             refactoring constants
# 2024-12-30: added support for lldp device id/name and cdp name from global device data to map L2 neighbour to CMK host
# 2025-01-01: added support for filter by host label/tag
#             INCOMPATIBLE:
#               cli options:
#                   changed:
#                       "--case" to "--l2-case"
#                       "--display-l2-neighbours" to "--l2-display-neighbours"
#                       "--include-l3-hosts" to "--l3-include-hosts"
#                       "--include-l3-loopback" to "--l3-include-loopback"
#                       "--keep" to "--keep-max-topologies"
#                       "--min-age" to "--min-topology-age"
#                       "--prefix" to "--l2-prefix"
#                       "--remove-domain" to "--l2-remove-domain"
#                       "--skip-l3-cidr-0" to "--l3-skip-cidr-0"
#                       "--skip-l3-cidr-32-128" to "--l3-skip-cidr-32-128"
#                       "--skip-l3-if" to "--l3-skip-if"
#                       "--skip-l3-ip" to "--l3-skip-ip"
#                       "--skip-l3-public" to "--l3-skip-public"
#                       "--check-user-data-only" ot "--check-config"
#                       "-u"/"--user-data-file" to "-c"/"-config"
#                   removed "-p" use "--l2-prefix" instead
#               TOML file:
#                   changed:
#                       "L2_HOST_MAP" to "L2_NEIGHBOUR_TO_HOST_MAP"
#                       "L3_REPLACE" to "L3_REPLACE_NETWORKS"
#                       "case" to "l2_case"
#                       "display_l2_neighbours" to "l2_display_neighbours"
#                       "include_l3_hosts" to "l3_include_hosts"
#                       "include_l3_loopback" to "l3_include_loopback"
#                       "keep" to "keep_max_topologies"
#                       "min-age" to "min_topology_age"
#                       "prefix" to "l2_prefix"
#                       "remove_domain" to "l2_remove_domain"
#                       "skip_l3_cidr_0" to "l3_skip_cidr_0"
#                       "skip_l3_cidr_32_128" to "l3_skip_cidr_32_128"
#                       "skip_l3_if" to "l3_skip_if"
#                       "skip_l3_ip" to "l3_skip_ip"
#                       "skip_l3_public" to "l3_skip_public"
# 2025-01-05: added "AUTO" to --l2-case parameters
#             INCOMPATIBLE: changed --l2-remove-domain from bool to "OFF" | "ON" | "AUTO"
#             added support for filter by folder
# 2024-01-06: added option "--l2-skip-external"
# 2024-01-07: added option "INSENSITIVE" to --l2-case
# 2025-01-11: INCOMPATIBLE: changed "--adjust-toml" -> "--update-config"
# 2025-01-18: INCOMPATIBLE: changed "CUSTOMERS = []" -> "FILTER_BY_CUSTOMER = []"
#                                   "SITES = []" -> "FILTER_BY_SITE = []"
# 2025-01-21: added support for Post requests (Werk #17003)
#             fixed REST API query for interface services
# 2025-01-24: added option --l2-display-ports, --l3-display-devices
# 2025-02-05: added option "OFF" to --l2-case
# 2025-02-07: added option --l2-ignore-mismatch
#
# creating topology data json from inventory data
#
# This script creates the topology data file needed for the Checkmk "network_visualization" plugin
# by Andreas Boesl/schnetz. For more information see
# https://forum.checkmk.com/t/network-visualization/41680  (from 2023-10-05)
# https://exchange.checkmk.com/p/network-visualization     (from 2023-10-05)
#
# NOTE: the topology_data configuration (layout etc.) is saved under ~/var/check_mk/topology/
#
# The inventory data could be created with my CDP/LLDP/IP Address/Interface nane inventory plugins:
# CDP.....: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_cdp_cache
# LLDP....: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_lldp_cache
# L3......: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_ip_addresses
#         : https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_lnx_if_ip
#         : https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_win_if_ip
# IF Name.: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_ifname
#
# USAGE:
# ~/local/lib/nvdct/nvdct.py --help
#

"""
# ##################################################################################################
# ##################################################################################################
# ##################################### data format ################################################
# ##################################################################################################
# ##################################################################################################

__objects = {
    "HostA": {
        "name": "myhost",
        "link": {"core": "myhost"},
        'metadata': {},
    },
    "Service@HostA": {
        "name": "SomeService",
        "link": {"core": ["myhost", "interface3"]},  # service@host in CMK
        # "link": {"core": "myhost"},                # host object in CMK
        # "link": {},                                # object not in CMK core
        'metadata': {
            'images': {
                'icon': 'icon_warning',              # additional image shown at the node
                'emblem': 'ip-network_80',           # image to show for an object
            },
            'tooltip': {
                'html': "Test message<div style='color: blue'>Some color <b>test</b></div>",
                'quickinfo': [
                                {'name': "Duplex", 'value': 'full duplex'},
                                {'name': 'Speed', 'value': '123132234'},
                                {'name': 'Native vlan', 'value': '123'}
                            ]
            }
        },
    },
}

__connections = [
    [["HostA", "ServiceOfHostA"], {}],
    [["ServiceOfHostA", "ServiceOfHostB"], {
        'line_config': {
            'thickness': 3,
            'color': '332211',
            'css_styles': {
                'stroke': 'green',
                'stroke-dasharray': '30 30',
                # 'stroke-dasharray': 'unset',  # to remove a CSS attribute
                'stroke-width': '4',
            }
        },
        'tooltip': {
            'html': "Test message mismatch<div style='color: blue'>Some color <b>test</b></div>",
            'quickinfo': [
                            {'name': "Duplex", 'value': 'full duplex'},
                            {'name': 'Native speed', 'value': '123132234'},
                            {'name': 'Native vlan', 'value': '123'}
                        ]
        }
    }],
    [["ServiceOfHostB", "HostB"], {}],
]

__data = {
    'version': 1,
    'name': 'LLDP',  # label for this layer
    'objects': __objects,
    'connections': __connections,
}

# ##################################################################################################
# ##################################################################################################
# ##################################### data format ################################################
# ##################################################################################################
# ##################################################################################################
"""

import sys

from collections.abc import MutableSequence
from time import strftime, time_ns

from typing import List

from lib.args import parse_arguments

from lib.update_config import update_config
from lib.backends import (
    HostCache,
    HostCacheLiveStatus,
    HostCacheMultiSite,
    HostCacheRestApi,
)
from lib.constants import (
    Backends,
    DATAPATH,
    ExitCodes,
    HostFilter,
    HostLabels,
    IPVersion,
    InvPaths,
    LOGGER,
    Layers,
    LiveStatusOperator,
    NVDCT_VERSION,
    TomlSections,
    TomlSettings,
    URLs,
)
from lib.settings import Settings
from lib.topologies import (
    TopologyL2,
    TopologyL3,
    TopologyStatic,
)
from lib.utils import (
    StdoutQuiet,
    configure_logger,
    remove_old_data,
)


def main():
    start_time = time_ns()

    settings: Settings = Settings(vars(parse_arguments()))

    sys.stdout = StdoutQuiet(quiet=settings.quiet)
    configure_logger(
        log_file=settings.log_file,
        log_to_console=settings.log_to_stdtout,
        log_level=settings.loglevel,
    )
    # always logg start and end of a session (except --log-level OFF)
    LOGGER.critical('Data creation started')

    print('')
    print(
        f'Network Visualisation Data Creation Tool (NVDCT)\n'
        f'by thl-cmk[at]outlook[dot]com, version {NVDCT_VERSION}\n'
        f'see {URLs.NVDCT}'
    )
    print('')
    print(f'Start time....: {strftime(settings.time_format)}')

    if settings.update_config:
        update_config(settings.config)
        print(f'Time taken....: {(time_ns() - start_time) / 1e9}/s')
        print(f'End time......: {strftime(settings.time_format)}')
        print('')

        LOGGER.critical('Data creation finished')
        sys.exit()

    match settings.backend:
        case Backends.MULTISITE:
            host_cache: HostCache = HostCacheMultiSite(
                customers=settings.filter_by_customer,
                filter_customers=settings.filter_customers,
                filter_sites=settings.filter_sites,
                pre_fetch=settings.pre_fetch,
                sites=settings.filter_by_site,
            )
        case Backends.RESTAPI:
            host_cache: HostCache = HostCacheRestApi(
                api_port=settings.api_port,
                filter_sites=settings.filter_sites,
                pre_fetch=settings.pre_fetch,
                sites=settings.filter_by_site,
            )
        case Backends.LIVESTATUS:
            host_cache: HostCache = HostCacheLiveStatus(
                pre_fetch=settings.pre_fetch,
            )
        case _:
            LOGGER.error(f'Backend {settings.backend} not implemented')
            host_cache: HostCache | None = None  # to keep linter happy
            sys.exit(ExitCodes.BACKEND_NOT_IMPLEMENTED)

    jobs: MutableSequence = []
    pre_fetch_layers: List[str] = []
    pre_fetch_host_list: List[str] = []

    for layer in settings.layers:
        match layer:
            case Layers.STATIC:
                jobs.append(layer)
            case Layers.L3V4:
                jobs.append(layer)
                host_cache.add_inventory_path(path=InvPaths.L3)
                pre_fetch_layers.append(HostLabels.L3V4_ROUTER)
            case Layers.CDP | Layers.LLDP:
                jobs.append(layer)
                host_cache.add_inventory_path(InvPaths.CDP if layer == Layers.CDP else InvPaths.LLDP)
                host_cache.add_inventory_path(InvPaths.CDP_GLOBAL if layer == Layers.CDP else InvPaths.LLDP_GLOBAL)
                pre_fetch_layers.append(HostLabels.CDP if layer == Layers.CDP else HostLabels.LLDP)
            case _:
                LOGGER.warning(f'Unknown layer {layer} dropped.')
                continue

    if not jobs:
        message = (
            f'No layer to work on. Please configura at least one layer (i.e. CLI option "-l {Layers.CDP}")\n'
            f'See {settings.config} -> {TomlSections.SETTINGS} -> {TomlSettings.LAYERS}'
        )
        LOGGER.warning(message)
        print(message)
        sys.exit(ExitCodes.NO_LAYER_CONFIGURED)

    # init filter lists before pre-fetch
    host_cache.init_filter_lists(
        filter_by_folder=settings.filter_by_folder,
        filter_by_host_label=settings.filter_by_host_label,
        filter_by_host_tag=settings.filter_by_host_tag,
    )

    if settings.pre_fetch:
        LOGGER.info('Pre fill cache...')
        for host_label in pre_fetch_layers:
            if host_list := host_cache.filter_host_list(
                    host_cache.query_hosts_by_filter(HostFilter.LABELS, host_label, LiveStatusOperator.EQUAL)
            ):
                pre_fetch_host_list = list(set(pre_fetch_host_list + list(host_list)))
        LOGGER.info(f'Fetching data for {len(pre_fetch_host_list)} hosts start')
        print(f'Prefetch start: {strftime(settings.time_format)}')
        print(f'Prefetch hosts: {len(pre_fetch_host_list)} of {len(host_cache.cache)}')
        host_cache.fill_cache(pre_fetch_host_list)
        LOGGER.info(f'Fetching data for {len(pre_fetch_host_list)} hosts end')
        print(f'Prefetch end..: {strftime(settings.time_format)}')

    # must not be used before pre-fetch is done
    host_cache.init_neighbour_to_host_map(
        case=settings.l2_case,
        l2_host_map=settings.l2_neighbour_to_host_map,
        prefix=settings.l2_prefix,
        remove_domain=settings.l2_remove_domain,
    )

    for job in jobs:
        match job:
            case Layers.STATIC:
                label = job
                topology = TopologyStatic(
                    connections=settings.static_connections,
                    emblems=settings.emblems,
                    host_cache=host_cache,
                )
            case Layers.L3V4:
                label = job
                topology = TopologyL3(
                    display_devices=settings.l3_display_devices,
                    emblems=settings.emblems,
                    host_cache=host_cache,
                    ignore_hosts=settings.l3_ignore_hosts,
                    ignore_ips=settings.l3_ignore_ips,
                    ignore_wildcard=settings.l3v4_ignore_wildcard,
                    include_hosts=settings.l3_include_hosts,
                    include_loopback=settings.l3_include_loopback,
                    replace=settings.l3_replace_networks,
                    skip_cidr_0=settings.l3_skip_cidr_0,
                    skip_cidr_32_128=settings.l3_skip_cidr_32_128,
                    skip_if=settings.l3_skip_if,
                    skip_ip=settings.l3_skip_ip,
                    skip_public=settings.l3_skip_public,
                    summarize=settings.l3_summarize,
                    version=IPVersion.IPv4 if job == Layers.L3V4 else IPVersion.IPv6,
                )
            case Layers.CDP | Layers.LLDP:
                label = job
                if job == Layers.CDP:
                    host_label = HostLabels.CDP
                    inv_path = InvPaths.CDP
                else:
                    host_label = HostLabels.LLDP
                    inv_path = InvPaths.LLDP
                if not (seed_devices := settings.l2_seed_devices):
                    seed_devices = host_cache.query_hosts_by_filter(
                        HostFilter.LABELS, host_label, LiveStatusOperator.EQUAL
                    )
                topology = TopologyL2(
                    display_neighbours=settings.l2_display_neighbours,
                    display_ports = settings.l2_display_ports,
                    drop_neighbours=settings.l2_drop_neighbours,
                    emblems=settings.emblems,
                    host_cache=host_cache,
                    label=label,
                    neighbour_replace_regex=settings.l2_neighbour_replace_regex,
                    path_in_inventory=inv_path,
                    seed_devices=seed_devices,
                    skip_external=settings.l2_skip_external,
                )
            case _:
                LOGGER.warning(f'Unknown layer {job}, ignoring.')
                continue
        pre_message = f'Layer {label:.<8s}: '
        print(pre_message, end ='', flush=True)

        topology.create()
        topology.nv_connections.add_meta_data_to_connections(
            nv_objects=topology.nv_objects,
            speed_map=settings.map_speed_to_thickness,
            ignore_mismatch=settings.l2_ignore_mismatch,
        )

        topology.save(
            label=label,
            output_directory=settings.output_directory,
            make_default=settings.default,
            dont_compare=settings.dont_compare,
        )

        message = (
            f'Devices/Objects/Connections added {topology.nv_objects.host_count}/'
            f'{len(topology.nv_objects.nv_objects)}/{len(topology.nv_connections.nv_connections)}'
        )
        LOGGER.info(f'{pre_message} {message}')
        print(message)

    if settings.keep_max_topologies:
        remove_old_data(
            keep=settings.keep_max_topologies,
            min_age=settings.min_topology_age,
            raw_path=DATAPATH,
            protected=settings.protected_topologies,
        )

    print(f'Time taken....: {(time_ns() - start_time) / 1e9}/s')
    print(f'End time......: {strftime(settings.time_format)}')
    print('')

    LOGGER.critical('Data creation finished')


if __name__ == '__main__':
    main()
